import unittest

from glossary_v2_test import TestGlossaryBase


class TestGlossaryFreeDict(TestGlossaryBase):
	def __init__(self, *args, **kwargs):
		TestGlossaryBase.__init__(self, *args, **kwargs)

		self.dataFileCRC32.update(
			{
				"100-en-de.tei": "542c210e",
				"100-en-de-v5.txt": "0c5942cd",
				"freedict-sample-2024-12-19.tei": "c33b89d5",
				"freedict-sample-2024-12-19-v2.txt": "c5eeff32",
			},
		)

	def convert_tei_txt(self, fname, fname2, **convertArgs):
		self.convert(
			f"{fname}.tei",
			f"{fname}-2.txt",
			compareText=f"{fname2}.txt",
			**convertArgs,
		)

	def test_convert_tei_txt_1(self):
		self.convert_tei_txt(
			"100-en-de",
			"100-en-de-v5",
			readOptions={"auto_comma": False},
		)
		self.convert_tei_txt(
			"100-en-de",
			"100-en-de-v5",
			readOptions={"auto_comma": True},
		)

	def test_convert_tei_txt_2(self):
		self.convert_tei_txt(
			"freedict-sample-2024-12-19",
			"freedict-sample-2024-12-19-v2",
		)


if __name__ == "__main__":
	unittest.main()



import sys
import unittest
from os.path import abspath, dirname

rootDir = dirname(dirname(abspath(__file__)))
sys.path.insert(0, rootDir)

from pyglossary.xml_utils import xml_escape


class Test_xml_escape(unittest.TestCase):
	def test(self):
		f = xml_escape
		self.assertEqual(f(""), "")
		self.assertEqual(f("abc"), "abc")
		self.assertEqual(f('"a"'), "&quot;a&quot;")
		self.assertEqual(f("'a'"), "&apos;a&apos;")
		self.assertEqual(f('"a"', quotation=False), '"a"')
		self.assertEqual(f("'a'", quotation=False), "'a'")
		self.assertEqual(f("R&D"), "R&amp;D")
		self.assertEqual(f("<-->"), "&lt;--&gt;")


if __name__ == "__main__":
	unittest.main()



import os
import unittest

from glossary_v2_test import TestGlossaryBase


class TestGlossarySlobWrite(TestGlossaryBase):
	def __init__(self, *args, **kwargs):
		TestGlossaryBase.__init__(self, *args, **kwargs)

		self.dataFileCRC32.update(
			{
				"300-ru-en.txt": "77cfee2f",
			},
		)

	def setUp(self):
		TestGlossaryBase.setUp(self)

	def test_convert_txt_slob_1(self):
		fname = "100-en-fa"
		os.environ["SLOB_TIMESTAMP"] = "2023-01-01T12:00:00.000000+00:00"
		self.convert(
			f"{fname}.txt",
			f"{fname}.slob",
			# sha1sum="",
			# compareBinary="",
			# slob file is different each time (and so its sha1sum and md5sum)
			# even with same exact tags!
			# writeOptions={"compression": ""},
		)

	def test_convert_txt_slob_2_file_size_approx(self):
		if os.getenv("SKIP_TEST_SLOB_WRITE_FILE_SIZE_APPROX"):
			print("skipping test_convert_txt_slob_2_file_size_approx")
			return
		fname = "300-ru-en"
		file_size_approx = 25000
		files = [
			(35852, self.newTempFilePath("300-ru-en.slob")),
			(35687, self.newTempFilePath("300-ru-en.1.slob")),
			(33856, self.newTempFilePath("300-ru-en.2.slob")),
			(29413, self.newTempFilePath("300-ru-en.3.slob")),
		]
		self.convert(
			f"{fname}.txt",
			f"{fname}.slob",
			writeOptions={
				"file_size_approx": file_size_approx,
				"file_size_approx_check_num_entries": 1,
			},
			compareBinary="",
			# slob file is different each time (and so its sha1sum and md5sum)
		)
		for size, fpath in files:
			with open(fpath, mode="rb") as file:
				actualSize = len(file.read())
			delta = actualSize - size
			self.assertLess(
				delta,
				100,
				msg=f"size expected={size} actual={actualSize}, file {fpath}",
			)


if __name__ == "__main__":
	unittest.main()



from typing import NotRequired, TypedDict

__all__ = ["ConfigType"]

# NotRequired is added in Python 3.11


class ConfigType(TypedDict):
	log_time: NotRequired[bool]
	cleanup: NotRequired[bool]
	auto_sqlite: NotRequired[bool]
	lower: NotRequired[bool]
	utf8_check: NotRequired[bool]
	enable_alts: NotRequired[bool]
	skip_resources: NotRequired[bool]
	rtl: NotRequired[bool]
	remove_html: NotRequired[str]
	remove_html_all: NotRequired[bool]
	normalize_html: NotRequired[bool]
	save_info_json: NotRequired[bool]
	skip_duplicate_headword: NotRequired[bool]
	trim_arabic_diacritics: NotRequired[bool]
	unescape_word_links: NotRequired[bool]
	# color.enable.cmd.unix: bool
	# color.enable.cmd.windows: bool
	# color.cmd.critical: int
	# color.cmd.error: int
	# color.cmd.warning: int
	# cmdi.prompt.indent.str: str
	# cmdi.prompt.indent.color: int
	# cmdi.prompt.msg.color: int
	# cmdi.msg.color: int
	# ui_autoSetFormat: bool
	# tk.progressbar.color.fill: str
	# tk.progressbar.color.background: str
	# tk.progressbar.color.text: str
	# tk.progressbar.font: str
	reverse_matchWord: NotRequired[bool]
	reverse_showRel: NotRequired[str]
	reverse_saveStep: NotRequired[int]
	reverse_minRel: NotRequired[float]
	reverse_maxNum: NotRequired[float]
	reverse_includeDefs: NotRequired[bool]



# -*- coding: utf-8 -*-
# mypy: ignore-errors
#
# Copyright © 2025 Saeed Rasooli <saeed.gnu@gmail.com> (ilius)
#
# This program is a free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
#
# You can get a copy of GNU General Public License along this program
# But you can always get it from http://www.gnu.org/licenses/gpl.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
from __future__ import annotations

import ast
import logging
import os
import tkinter as tk
import traceback
from os.path import isfile, join, splitext
from tkinter import filedialog, ttk
from tkinter import font as tkFont
from typing import TYPE_CHECKING, Any, Protocol

from pyglossary.core import confDir, homeDir, homePage, sysName
from pyglossary.glossary_v2 import ConvertArgs, Error, Glossary
from pyglossary.option import IntOption, Option
from pyglossary.os_utils import abspath2
from pyglossary.sort_keys import namedSortKeyList
from pyglossary.text_utils import escapeNTB, unescapeNTB, urlToPath

from .base import (
	UIBase,
	aboutText,
	authors,
	licenseText,
	logo,
)
from .config import configDefDict
from .version import getVersion

if TYPE_CHECKING:
	from collections.abc import Callable
	from tkinter.font import Font

	from pyglossary.config_type import ConfigType
	from pyglossary.logger import Logger

__all__ = ["UI"]

log: Logger = logging.getLogger("pyglossary")

# on Windows: make app DPI-aware, fix blurry fonts
# this causes issues on tk 8.6.x with custom fonts or display scaling ≠ %100
# because many widgets worked based on an assumed system font metric that
# doesn’t always match your actual font
# see https://github.com/ilius/pyglossary/issues/689
if os.sep == "\\" and tk.TkVersion >= 8.7:
	import ctypes

	try:
		# 1 = system DPI aware
		ctypes.windll.shcore.SetProcessDpiAwareness(1)
	except Exception:
		log.exception("")


pluginByDesc = {plugin.description: plugin for plugin in Glossary.plugins.values()}
readDesc = [plugin.description for plugin in Glossary.plugins.values() if plugin.canRead]
writeDesc = [
	plugin.description for plugin in Glossary.plugins.values() if plugin.canWrite
]


def set_window_icon(window: tk.Toplevel) -> None:
	window.iconphoto(
		True,
		tk.PhotoImage(file=logo),
	)


def decodeGeometry(gs: str) -> tuple[int, int, int, int]:
	"""
	Example for gs: "253x252+30+684"
	returns (x, y, w, h).
	"""
	p = gs.split("+")
	w, h = p[0].split("x")
	return (int(p[1]), int(p[2]), int(w), int(h))


def encodeGeometry(x: int, y: int, w: int, h: int) -> str:
	return f"{w}x{h}+{x}+{y}"


def encodeLocation(x: int, y: int) -> str:
	return f"+{x}+{y}"


def centerWindow(win: tk.Tk) -> None:
	"""
	Centers a tkinter window
	:param win: the root or Toplevel window to center.
	"""
	win.update_idletasks()
	width = win.winfo_width()
	frm_width = win.winfo_rootx() - win.winfo_x()
	win_width = width + 2 * frm_width
	height = win.winfo_height()
	titlebar_height = win.winfo_rooty() - win.winfo_y()
	win_height = height + titlebar_height + frm_width
	x = win.winfo_screenwidth() // 2 - win_width // 2
	y = win.winfo_screenheight() // 2 - win_height // 2
	win.geometry(encodeGeometry(x, y, width, height))
	win.deiconify()


def newButton(
	master: tk.Widget,
	text: str,
	command: Callable,
	**kwargs: Any,
) -> ttk.Button:
	button = ttk.Button(
		master,
		text=text,
		command=command,
		**kwargs,
	)

	def onEnter(_event: tk.Event) -> None:
		button.invoke()

	button.bind("<Return>", onEnter)
	button.bind("<KP_Enter>", onEnter)
	return button


def newLabelWithImage(parent: tk.Widget, file: str = "") -> ttk.Label:
	image = tk.PhotoImage(file=file)
	label = ttk.Label(parent, image=image)
	# keep a reference:
	label.image = image
	return label


def newReadOnlyText(
	parent: tk.Widget,
	text: str = "",
	borderwidth: int = 10,
	font: tuple[str, int, str] = ("DejaVu Sans", 11, ""),
) -> tk.Text:
	height = len(text.strip().split("\n"))
	widget = tk.Text(
		parent,
		height=height,
		borderwidth=borderwidth,
		font=font,
	)
	widget.insert(1.0, text)
	widget.pack()

	# widget.bind("<Key>", lambda e: break)
	widget.configure(state=tk.DISABLED)

	return widget


class TkTextLogHandler(logging.Handler):
	def __init__(self, tktext: tk.Text) -> None:
		logging.Handler.__init__(self)
		#####
		tktext.tag_config("CRITICAL", foreground="#ff0000")
		tktext.tag_config("ERROR", foreground="#ff0000")
		tktext.tag_config("WARNING", foreground="#ffff00")
		tktext.tag_config("INFO", foreground="#00ff00")
		tktext.tag_config("DEBUG", foreground="#ffffff")
		tktext.tag_config("TRACE", foreground="#ffffff")
		###
		self.tktext = tktext

	def emit(self, record: logging.LogRecord) -> None:
		msg = ""
		if record.getMessage():
			msg = self.format(record)
		###
		if record.exc_info:
			type_, value, tback = record.exc_info
			tback_text = "".join(
				traceback.format_exception(type_, value, tback),
			)
			if msg:
				msg += "\n"
			msg += tback_text
		###
		self.tktext.insert(
			"end",
			msg + "\n",
			record.levelname,
		)


# Monkey-patch Tkinter
# http://stackoverflow.com/questions/5191830/python-exception-logging
def CallWrapper__call__(self: tk.CallWrapper, *args: str) -> Any:
	"""Apply first function SUBST to arguments, than FUNC."""
	if self.subst:
		args = self.subst(*args)
	try:
		return self.func(*args)
	except Exception:
		log.exception("Exception in Tkinter callback:")


tk.CallWrapper.__call__ = CallWrapper__call__


class ProgressBar(ttk.Frame):
	def __init__(  # noqa: PLR0913
		self,
		rootWin: tk.Tk,
		min_: float,
		max_: float,
		width: int,
		height: int,
		appearance: str,  # "sunken"
		fillColor: str,
		background: str,
		labelColor: str,
		labelFont: str,
		value: float = 0,
	) -> None:
		self.min = min_
		self.max = max_
		self.width = width
		self.height = height
		self.value = value
		ttk.Frame.__init__(
			self,
			rootWin,
			relief=appearance,
		)
		self.canvas = tk.Canvas(
			self,
			height=height,
			width=width,
			bd=0,
			highlightthickness=0,
			background=background,
		)
		self.scale = self.canvas.create_rectangle(
			0,
			0,
			width,
			height,
			fill=fillColor,
		)
		self.label = self.canvas.create_text(
			width / 2,
			height / 2,
			text="",
			anchor="center",
			fill=labelColor,
			font=labelFont,
		)
		self.update()
		self.bind("<Configure>", self.update)
		self.canvas.pack(side="top", fill="x", expand=False)

	def updateProgress(
		self,
		value: float,
		max_: float | None = None,
		text: str = "",
	) -> None:
		if max_:
			self.max = max_
		self.value = value
		self.update(None, text)

	def update(self, event: tk.Event | None = None, labelText: str = "") -> None:
		if event:  # instance of tkinter.Event
			width = getattr(event, "width", None) or int(self.winfo_width())
			if width != self.width:  # window is resized
				self.canvas.coords(self.label, width / 2, self.height / 2)
				self.width = width
		else:
			width = self.width

		self.canvas.coords(
			self.scale,
			0,
			0,
			width * max(min(self.value, self.max), self.min) / self.max,
			self.height,
		)

		if labelText:
			# TODO: replace below `// 10` with a number based on current font size
			self.canvas.itemconfig(
				self.label,
				text=labelText[: width // 10],
			)

		self.canvas.update_idletasks()


# class VerticalProgressBar(ProgressBar):
# def update(self, event=None, labelText="") -> None:
# ...
# self.canvas.coords(
# 	self.scale,
# 	0,
# 	self.height * (1 - value / self.max),
# 	width,
# 	self.height,
# )


class FormatDialog(tk.Toplevel):
	def __init__(  # noqa: PLR0913
		self,
		rootWin: tk.Tk,
		descList: list[str],
		title: str,
		onOk: Callable,
		button: FormatButton,
		activeDesc: str = "",
	) -> None:
		tk.Toplevel.__init__(self)
		# bg="#0f0" does not work
		self.descList = descList
		self.items = self.descList
		self.onOk = onOk
		self.activeDesc = activeDesc
		self.lastSearch = None
		self.resizable(width=True, height=True)
		if title:
			self.title(title)
		set_window_icon(self)
		self.bind("<Escape>", lambda _e: self.destroy())

		px, py, pw, ph = decodeGeometry(button.winfo_toplevel().geometry())
		width = 400
		height = 400
		self.geometry(
			encodeGeometry(
				px + pw // 2 - width // 2,
				py + ph // 2 - height // 2,
				width,
				height,
			),
		)

		entryBox = ttk.Frame(master=self)
		ttk.Label(master=entryBox, text="Search: ").pack(side="left")
		entry = self.entry = ttk.Entry(master=entryBox)
		entry.pack(fill="x", expand=True, side="left")
		entryBox.pack(fill="x", padx=5, pady=5)

		entry.bind("<KeyRelease>", self.onEntryKeyRelease)
		entry.focus()

		treevBox = ttk.Frame(master=self)

		treev = self.treev = ttk.Treeview(
			master=treevBox,
			columns=["Description"],
			show="",
		)
		treev.bind("<Double-1>", self.onTreeDoubleClick)
		treev.pack(
			side="left",
			fill="both",
			expand=True,
		)

		vsb = ttk.Scrollbar(
			master=treevBox,
			orient="vertical",
			command=treev.yview,
		)
		vsb.pack(side="right", fill="y")

		treevBox.pack(
			fill="both",
			expand=True,
			padx=5,
			pady=5,
		)

		treev.configure(yscrollcommand=vsb.set)

		if rootWin.tk.call("tk", "windowingsystem") == "x11":
			treev.bind("<Button-4>", self.onTreeviewMouseWheel)
			treev.bind("<Button-5>", self.onTreeviewMouseWheel)

		self.updateTree()

		buttonBox = ttk.Frame(master=self)

		cancelButton = newButton(
			buttonBox,
			text="Cancel",
			command=self.cancelClicked,
		)
		cancelButton.pack(side="right")

		okButton = newButton(
			buttonBox,
			text="  OK  ",
			command=self.okClicked,
			# bg="#ff0000",
			# activebackground="#ff5050",
		)
		okButton.pack(side="right")

		buttonBox.pack(fill="x")

		self.bind("<Return>", self.onReturnPress)
		self.bind("<KP_Enter>", self.onReturnPress)

		self.bind("<Down>", self.onDownPress)
		self.bind("<Up>", self.onUpPress)

		# self.bind("<KeyPress>", self.onKeyPress)

	def onTreeviewMouseWheel(self, event: tk.Event) -> str | None:
		# only register this on X11 (Linux / BSD)
		if not hasattr(event, "num"):
			return None

		self.treev.yview_scroll(-1 if event.num == 4 else 1, "units")
		return "break"

	def setActiveRow(self, desc: str) -> None:
		self.treev.selection_set(desc)
		self.treev.see(desc)

	def updateTree(self) -> None:
		treev = self.treev
		current = treev.get_children()
		if current:
			treev.delete(*current)
		for desc in self.items:
			treev.insert("", "end", values=[desc], iid=desc)
			# iid should be rowId
		if self.activeDesc in self.items:
			self.setActiveRow(self.activeDesc)

	def onEntryKeyRelease(self, _event: tk.Event) -> None:
		text = self.entry.get().strip()
		if text == self.lastSearch:
			return

		if not text:
			self.items = self.descList
			self.updateTree()
			self.lastSearch = text
			return

		text = text.lower()
		descList = self.descList

		items1 = []
		items2 = []
		for desc in descList:
			if desc.lower().startswith(text):
				items1.append(desc)
			elif text in desc.lower():
				items2.append(desc)

		self.items = items1 + items2
		self.updateTree()
		self.lastSearch = text

	def onTreeDoubleClick(self, _event: tk.Event) -> None:
		self.okClicked()

	def cancelClicked(self) -> None:
		self.destroy()

	def onReturnPress(self, _event: tk.Event) -> None:
		self.okClicked()

	def onDownPress(self, _event: tk.Event) -> None:
		treev = self.treev
		selection = treev.selection()
		if selection:
			nextDesc = treev.next(selection[0])
			if nextDesc:
				self.setActiveRow(nextDesc)
		elif self.items:
			self.setActiveRow(self.items[0])
		treev.focus()

	def onUpPress(self, _event: tk.Event) -> None:
		treev = self.treev
		treev.focus()
		selection = treev.selection()
		if not selection:
			if self.items:
				self.setActiveRow(self.items[0])
			return
		nextDesc = treev.prev(selection[0])
		if nextDesc:
			self.setActiveRow(nextDesc)

	def onKeyPress(self, event: tk.Event) -> None:
		print(f"FormatDialog: onKeyPress: {event}")

	def okClicked(self) -> None:
		treev = self.treev
		selectedList = treev.selection()
		desc = selectedList[0] if selectedList else ""
		self.onOk(desc)
		self.destroy()


class FormatButton(ttk.Button):
	noneLabel = "[Select Format]"

	def __init__(
		self,
		rootWin: tk.Tk,
		descList: list[str],
		dialogTitle: str,
		onChange: Callable,
		master: tk.Widget | None = None,
	) -> None:
		self.rootWin = rootWin
		self.var = tk.StringVar()
		self.var.set(self.noneLabel)
		ttk.Button.__init__(
			self,
			master=master,
			textvariable=self.var,
			command=self.onClick,
		)
		self.descList = descList
		self.dialogTitle = dialogTitle
		self._onChange = onChange
		self.activeDesc = ""
		self.bind("<Return>", self.onEnter)
		self.bind("<KP_Enter>", self.onEnter)

	def onEnter(self, _event: tk.Event | None = None) -> None:
		self.invoke()

	def onChange(self, desc: str) -> None:
		self.setValue(desc)
		self._onChange(desc)

	def get(self) -> str:
		return self.activeDesc

	def setValue(self, desc: str) -> None:
		if desc:
			self.var.set(desc)
		else:
			self.var.set(self.noneLabel)
		self.activeDesc = desc

	def onClick(self) -> None:
		dialog = FormatDialog(
			self.rootWin,
			descList=self.descList,
			title=self.dialogTitle,
			onOk=self.onChange,
			button=self,
			activeDesc=self.activeDesc,
		)
		dialog.focus()


class OptionTkType(Protocol):
	def __init__(self, opt: Option, parent: ttk.Widget) -> None: ...
	@property
	def value(self) -> Any: ...
	@value.setter
	def value(self, x: Any) -> None: ...
	@property
	def widget(self) -> ttk.Widget: ...


class BoolOptionTk:
	def __init__(self, opt: Option, parent: ttk.Widget) -> None:
		frame = ttk.Frame(master=parent)
		frame.pack(side="top", fill="y", expand=True)
		var = tk.IntVar()
		cb = ttk.Checkbutton(
			master=frame,
			variable=var,
			text=f"{opt.displayName} ({opt.comment})",
		)
		cb.pack(fill="x")
		self._frame = frame
		self._var = var

	@property
	def value(self) -> Any:
		return bool(self._var.get())

	@value.setter
	def value(self, x: Any) -> None:
		self._var.set(int(bool(x)))

	@property
	def widget(self) -> ttk.Widget:
		return self._frame


class IntOptionTk:
	def __init__(self, opt: Option, parent: ttk.Widget) -> None:
		assert isinstance(opt, IntOption)
		frame = ttk.Frame(master=parent)
		frame.pack(side="top", fill="y", expand=True)
		ttk.Label(
			master=frame,
			text=f"{opt.displayName}: ",
		).pack(
			side="left",
			expand=False,
		)
		minim = opt.minim
		if minim is None:
			minim = 0
		maxim = opt.maxim
		if maxim is None:
			maxim = 1_000_000_000
		spin = ttk.Spinbox(
			master=frame,
			from_=minim,
			to=maxim,
			increment=1,
			width=max(len(str(minim)), len(str(maxim))) + 1,
		)
		spin.pack(side="left", expand=False)
		ttk.Label(
			master=frame,
			text=opt.comment,
		).pack(
			side="left",
			fill="x",
			expand=True,
		)
		self._frame = frame
		self._spin = spin

	@property
	def value(self) -> Any:
		return int(self._spin.get())

	@value.setter
	def value(self, x: Any) -> None:
		self._spin.set(int(x))

	@property
	def widget(self) -> ttk.Widget:
		return self._frame


class StrOptionTk:
	def __init__(self, opt: Option, parent: ttk.Widget) -> None:
		self.opt = opt
		frame = ttk.Frame(master=parent)
		frame.pack(side="top", fill="y", expand=True)
		ttk.Label(
			master=frame,
			text=f"{opt.displayName} ({opt.comment}): ",
		).pack(side="left")
		entry = ttk.Entry(master=frame)
		entry.pack(side="left", fill="x", expand=True)
		self._frame = frame
		self._entry = entry

		self.valuesVar = None
		if opt.values:
			var = self.valuesVar = tk.StringVar()
			var.trace_add("write", self._optionMenuActivated)
			optMenu = ttk.OptionMenu(frame, var, *tuple(opt.values))
			optMenu.pack(side="left", fill="x")

	def _optionMenuActivated(self, *_args: Any) -> None:
		self.value = self.valuesVar.get()

	@property
	def value(self) -> Any:
		return self.opt.evaluate(self._entry.get())[0]

	@value.setter
	def value(self, x: Any) -> None:
		x = str(x)
		if self.valuesVar is not None:
			self.valuesVar.set(x)
		self._entry.delete(0, tk.END)
		self._entry.insert(0, x)

	@property
	def widget(self) -> ttk.Widget:
		return self._frame


class FileSizeOptionTk(StrOptionTk):
	pass


class HtmlColorOptionTk(StrOptionTk):
	pass


class MultiLineStrOptionTk:
	def __init__(self, opt: Option, parent: ttk.Widget) -> None:
		frame = ttk.Frame(master=parent)
		frame.pack(side="top", fill="y", expand=True)
		ttk.Label(
			master=frame, text=f"{opt.displayName}: {opt.comment} (escaped \\n\\t): "
		).pack(side="left")
		entry = ttk.Entry(master=frame)
		entry.pack(side="left", fill="x", expand=True)
		self._frame = frame
		self._entry = entry

	@property
	def value(self) -> Any:
		return unescapeNTB(self._entry.get())

	@value.setter
	def value(self, x: Any) -> None:
		self._entry.insert(0, escapeNTB(str(x)))

	@property
	def widget(self) -> ttk.Widget:
		return self._frame


class NewlineOptionTk(MultiLineStrOptionTk):
	pass


class LiteralEvalOptionTk:
	typeHint = ""

	def __init__(self, opt: Option, parent: ttk.Widget) -> None:
		frame = ttk.Frame(master=parent)
		frame.pack(side="top", fill="y", expand=True)
		ttk.Label(
			master=frame,
			text=f"{opt.displayName} ({opt.comment}, {self.typeHint}): ",
		).pack(side="left")
		entry = ttk.Entry(master=frame)
		entry.pack(side="left", fill="x", expand=True)
		self._frame = frame
		self._entry = entry

	@property
	def value(self) -> Any:
		return ast.literal_eval(self._entry.get())

	@value.setter
	def value(self, x: Any) -> None:
		self._entry.insert(0, repr(x))

	@property
	def widget(self) -> ttk.Widget:
		return self._frame


class ListOptionTk(LiteralEvalOptionTk):
	typeHint = "Python list"


class DictOptionTk(LiteralEvalOptionTk):
	typeHint = "Python dict"


optionClassByName: dict[str, OptionTkType] = {
	"BoolOption": BoolOptionTk,
	"IntOption": IntOptionTk,
	"StrOption": StrOptionTk,
	"EncodingOption": StrOptionTk,
	"FileSizeOption": FileSizeOptionTk,
	"UnicodeErrorsOption": StrOptionTk,
	"HtmlColorOption": HtmlColorOptionTk,
	"NewlineOption": NewlineOptionTk,
	"ListOption": ListOptionTk,
	"DictOption": DictOptionTk,
	# "FloatOption": FloatOptionTk,  # not used so far!
}


class FormatOptionsDialog(tk.Toplevel):
	commentLen = 60
	kindFormatsOptions = {
		"Read": Glossary.formatsReadOptions,
		"Write": Glossary.formatsWriteOptions,
	}

	def __init__(
		self,
		formatDesc: str,
		kind: str,  # "Read" or "Write"
		values: dict[str, Any],
		okFunc: Callable[[dict[str, Any]]],
		master: tk.Widget | None = None,  # noqa: ARG002
	) -> None:
		formatName = pluginByDesc[formatDesc].name
		tk.Toplevel.__init__(self, master=master)
		# bg="#0f0" does not work
		self.resizable(width=True, height=True)
		self.title(f"{formatDesc} {kind} Options")
		set_window_icon(self)
		self.bind("<Escape>", lambda _e: self.destroy())

		self.format = formatName
		self.kind = kind
		self.values = values
		self.okFunc = okFunc
		self.options = self.kindFormatsOptions[kind][formatName]
		self.optionsProp = Glossary.plugins[formatName].optionsProp
		self.widgets: dict[str, OptionTkType] = {}

		self.createOptionsList()

		buttonBox = ttk.Frame(self)
		okButton = newButton(
			buttonBox,
			text="  OK  ",
			command=self.okClicked,
			# bg="#ff0000",
			# activebackground="#ff5050",
		)
		okButton.pack(side="right")
		buttonBox.pack(fill="x")

	def createOptionsList(self) -> None:
		values = self.values
		frame = self.frame = ttk.Frame(master=self)
		frame.pack(fill="x", expand=True)
		for optName, default in self.options.items():
			prop = self.optionsProp[optName]
			comment = prop.longComment
			if len(comment) > self.commentLen:
				comment = comment[: self.commentLen] + "..."
			widgetClass = optionClassByName.get(prop.__class__.__name__)
			if widgetClass is None:
				log.warning(f"No widget class for option class {prop.__class__}")
				continue
			w = widgetClass(prop, frame)
			ww = w.widget
			ww.parent = frame
			ww.pack(fill="x")
			w.value = values.get(optName, default)
			self.widgets[optName] = w

	def okClicked(self) -> None:
		for optName, widget in self.widgets.items():
			self.values[optName] = widget.value
		self.okFunc(self.values)
		self.destroy()


sortKeyNames = [sk.name for sk in namedSortKeyList]
sortKeyNameByDesc = {sk.desc: sk.name for sk in namedSortKeyList}
sortKeyDescByName = {sk.name: sk.desc for sk in namedSortKeyList}


class SortOptionsBox(ttk.Frame):
	def __init__(
		self,
		ui: UI,
		master: tk.Widget | None = None,
	) -> None:
		self.ui = ui
		ttk.Frame.__init__(self, master=master)
		hbox = ttk.Frame(master=self)
		sortCheckVar = tk.IntVar()
		sortCheck = ttk.Checkbutton(
			master=hbox,
			text="Sort entries by",
			variable=sortCheckVar,
			command=self.onSortCheckClicked,
		)
		sortKeyVar = tk.StringVar()
		sortKeyDescList = [sk.desc for sk in namedSortKeyList]
		sortKeyCombo = ttk.OptionMenu(
			hbox,
			sortKeyVar,
			sortKeyDescList[0],
			*sortKeyDescList,
		)
		self.sortCheckVar = sortCheckVar
		self.sortKeyVar = sortKeyVar
		self.sortKeyCombo = sortKeyCombo
		sortCheck.pack(side="left")
		sortKeyCombo.pack(side="left")
		hbox.pack(side="top", expand=True, fill="both")
		###
		hbox = self.encodingHBox = ttk.Frame(master=self)
		encodingCheckVar = self.encodingCheckVar = tk.IntVar()
		encodingCheck = ttk.Checkbutton(
			master=hbox,
			text="Sort Encoding",
			variable=encodingCheckVar,
		)
		encodingEntry = self.encodingEntry = ttk.Entry(hbox, width=15)
		encodingEntry.delete(0, "end")
		encodingEntry.insert(0, "utf-8")
		ttk.Label(hbox, text="    ").pack(side="left")
		encodingCheck.pack(side="left")
		encodingEntry.pack(side="left")
		hbox.pack(side="top", expand=True, fill="both")
		###
		hbox = self.localeHBox = ttk.Frame(master=self)
		localeEntry = self.localeEntry = ttk.Entry(hbox, width=15)
		ttk.Label(master=hbox, text="    ").pack(side="left")
		ttk.Label(master=hbox, text="Sort Locale").pack(side="left")
		localeEntry.pack(side="left", expand=True, fill="x")
		hbox.pack(side="top", expand=True, fill="both")
		###

	def updateSortStates(self, sort: bool) -> None:
		state = tk.NORMAL if sort else tk.DISABLED
		self.sortKeyCombo.configure(state=state)
		# self.encodingHBox.configure(state=state) # unknown option "-state"
		# self.localeHBox.configure(state=state) # unknown option "-state"

	def onSortCheckClicked(self) -> None:
		sort = bool(self.sortCheckVar.get())
		self.updateSortStates(sort)

	def updateWidgets(self) -> None:
		convertOptions = self.ui.convertOptions
		sort = bool(convertOptions.get("sort", False))
		self.sortCheckVar.set(int(sort))
		self.updateSortStates(sort)

		sortKeyName = convertOptions.get("sortKeyName")
		if sortKeyName:
			sortKeyName, _, localeName = sortKeyName.partition(":")
			if sortKeyName:
				self.sortKeyVar.set(sortKeyDescByName[sortKeyName])
			self.localeEntry.delete(0, "end")
			self.localeEntry.insert(0, localeName)

		if "sortEncoding" in convertOptions:
			self.encodingCheckVar.set(1)
			self.encodingEntry.delete(0, "end")
			self.encodingEntry.insert(0, convertOptions["sortEncoding"])

	def applyChanges(self) -> None:
		convertOptions = self.ui.convertOptions
		sort = int(self.sortCheckVar.get())
		if not sort:
			for param in ("sort", "sortKeyName", "sortEncoding"):
				if param in convertOptions:
					del convertOptions[param]
			return

		sortKeyDesc = self.sortKeyVar.get()
		sortKeyName = sortKeyNameByDesc[sortKeyDesc]
		sortLocale = self.localeEntry.get()
		if sortLocale:
			sortKeyName = f"{sortKeyName}:{sortLocale}"

		convertOptions["sort"] = True
		convertOptions["sortKeyName"] = sortKeyName
		if self.encodingCheckVar.get():
			convertOptions["sortEncoding"] = self.encodingEntry.get()


class GeneralOptionsDialog(tk.Toplevel):
	def __init__(
		self,
		ui: UI,
		master: tk.Widget | None = None,
	) -> None:
		self.ui = ui
		tk.Toplevel.__init__(self, master=master)
		self.resizable(width=True, height=True)
		self.title("General Options")
		set_window_icon(self)
		self.bind("<Escape>", lambda _e: self.destroy())
		####
		padx = 0
		pady = 0
		##
		self.sortOptionsBox = SortOptionsBox(ui, master=self)
		self.sortOptionsBox.pack(side="top", pady=pady, expand=True, fill="both")
		##
		hbox = ttk.Frame(self)
		self.sqliteCheckVar = tk.IntVar()
		self.sqliteCheck = ttk.Checkbutton(
			hbox,
			text="SQLite mode",
			variable=self.sqliteCheckVar,
		)
		self.sqliteCheck.pack(side="left", padx=padx, expand=True, fill="x")
		hbox.pack(side="top", pady=pady, expand=True, fill="both")
		##
		self.configParams = {
			"save_info_json": False,
			"lower": False,
			"skip_resources": False,
			"rtl": False,
			"enable_alts": True,
			"cleanup": True,
			"remove_html_all": True,
		}
		self.configCheckVars: dict[str, tk.IntVar] = {}
		for param in self.configParams:
			# hbox = ttk.Frame(self)
			comment = configDefDict[param].comment
			comment = comment.split("\n")[0]
			checkVar = tk.IntVar()
			checkButton = ttk.Checkbutton(
				self,
				text=comment,
				variable=checkVar,
			)
			self.configCheckVars[param] = checkVar
			checkButton.pack(side="top", padx=padx, expand=True, fill="both")
			# hbox.pack(side="top", pady=pady)

		###
		buttonBox = ttk.Frame(self)
		okButton = newButton(
			buttonBox,
			text="  OK  ",
			command=self.okClicked,
		)
		okButton.pack(side="right")
		buttonBox.pack(fill="x")
		##
		self.updateWidgets()

	def getSQLite(self) -> bool:
		convertOptions = self.ui.convertOptions
		sqlite = convertOptions.get("sqlite")
		if sqlite is not None:
			return sqlite
		return self.ui.config.get("auto_sqlite", True)

	def updateWidgets(self) -> None:
		config = self.ui.config
		self.sortOptionsBox.updateWidgets()
		self.sqliteCheck.configure(state=tk.NORMAL if self.getSQLite() else tk.DISABLED)
		for param, checkVar in self.configCheckVars.items():
			default = self.configParams[param]
			checkVar.set(config.get(param, default))

	def applyChanges(self) -> None:
		# print("applyChanges")
		self.sortOptionsBox.applyChanges()

		convertOptions = self.ui.convertOptions
		config = self.ui.config

		convertOptions["sqlite"] = bool(self.sqliteCheckVar.get())

		for param, checkVar in self.configCheckVars.items():
			config[param] = bool(checkVar.get())

	def okClicked(self) -> None:
		self.applyChanges()
		self.destroy()


class PreConvertInfoDialog(tk.Toplevel):
	def __init__(
		self,
		info: dict[str, Any],
		okFunc: Callable[[dict[str, Any]]],
		master: tk.Widget,
	) -> None:
		tk.Toplevel.__init__(self, master=master)
		self.resizable(width=True, height=True)
		self.title("Set Info / Metedata")
		set_window_icon(self)
		self.bind("<Escape>", lambda _e: self.destroy())

		self.info = info
		self.okFunc = okFunc

		mainFrame = self.frame = ttk.Frame(master=self)

		frame = ttk.Frame(master=mainFrame)
		ttk.Label(
			text="Glossary Name",
			master=frame,
		).pack(side="left", expand=False)
		nameEntry = ttk.Entry(master=frame, width=50)
		nameEntry.pack(side="left", fill="x", expand=True)
		nameEntry.insert(0, info.get("name", ""))
		self.nameEntry = nameEntry
		frame.pack(side="top", expand=True, fill="both")

		frame = ttk.Frame(master=mainFrame)
		ttk.Label(
			text="Source Language",
			master=frame,
		).pack(side="left", expand=False)
		sourceLangEntry = ttk.Entry(master=frame)
		sourceLangEntry.pack(side="right", fill="x", expand=True)
		sourceLangEntry.insert(0, info.get("sourceLang", ""))
		self.sourceLangEntry = sourceLangEntry
		frame.pack(side="top", expand=True, fill="both")

		frame = ttk.Frame(master=mainFrame)
		ttk.Label(
			text="Target Language",
			master=frame,
		).pack(side="left", expand=False)
		targetLangEntry = ttk.Entry(master=frame)
		targetLangEntry.pack(side="right", fill="x", expand=True)
		targetLangEntry.insert(0, info.get("targetLang", ""))
		self.targetLangEntry = targetLangEntry
		frame.pack(side="top", expand=True, fill="both")

		mainFrame.pack(fill="both", expand=True)

		buttonBox = ttk.Frame(self)
		okButton = newButton(
			buttonBox,
			text="  OK  ",
			command=self.okClicked,
			# bg="#ff0000",
			# activebackground="#ff5050",
		)
		okButton.pack(side="right")
		buttonBox.pack(fill="x")

		px, py, pw, ph = decodeGeometry(master.winfo_toplevel().geometry())
		self.geometry(
			encodeLocation(
				px + pw // 2 - 200,
				py + ph // 2 - 100,
			)
		)

	def okClicked(self) -> None:
		name = self.nameEntry.get()
		if name:
			self.info["name"] = name
		sourceLang = self.sourceLangEntry.get()
		if sourceLang:
			self.info["sourceLang"] = sourceLang
		targetLang = self.targetLangEntry.get()
		if targetLang:
			self.info["targetLang"] = targetLang
		self.okFunc(self.info)
		self.destroy()


class VerticalNotebook(ttk.Frame):
	def __init__(
		self,
		parent: tk.Widget,
		font: Font,
		**kwargs: Any,
	) -> None:
		ttk.Frame.__init__(self, parent, **kwargs)
		self.rowconfigure(0, weight=1)
		self.columnconfigure(2, weight=1)
		# scrollable tabs
		self._listbox = tk.Listbox(
			self,
			width=1,
			highlightthickness=0,
			relief="raised",
			justify="center",
			font=font,
		)
		self._listbox.configure()

		# list of widgets associated with the tabs
		self._tabs = []
		self._current_tab = None  # currently displayed tab

		self._listbox.grid(row=0, column=1, sticky="ns")
		# binding to display the selected tab
		self._listbox.bind("<<ListboxSelect>>", self._on_listbox_select)
		self._maxWidth = 0

	# add tab
	def add(self, widget: tk.Widget, text: str) -> None:
		self._listbox.insert("end", text)
		# resize listbox to be large enough to show all tab labels
		self._maxWidth = max(self._maxWidth, len(text))
		self._listbox.configure(
			width=self._maxWidth + 2,
		)
		index = len(self._tabs)
		self._tabs.append(widget)
		if self._current_tab is None:
			self.switch_tab(index)

	def switch_tab(self, index: int) -> None:
		self._show_tab_index(index)
		self._listbox.selection_clear(0, "end")
		self._listbox.selection_set(index)
		self._listbox.see(index)

	def _show_tab_index(self, index: int) -> None:
		widget = self._tabs[index]
		if self._current_tab is not None:
			self._current_tab.grid_remove()
		self._current_tab = widget
		widget.grid(in_=self, column=2, row=0, sticky="ewns")

	def _on_listbox_select(self, _event: tk.Event | None = None) -> None:
		selection = self._listbox.curselection()
		if not selection:
			return
		index = selection[0]
		if index >= len(self._tabs):
			print(f"{index=}")
			return
		self._show_tab_index(index)


class UI(tk.Frame, UIBase):
	fcd_dir_save_path = join(confDir, "ui-tk-fcd-dir")

	def __init__(
		self,
		progressbar: bool = True,
	) -> None:
		rootWin = self.rootWin = tk.Tk(
			# baseName="PyGlossary",  # no effect
			className="PyGlossary",
			# ^ capitalized as Pyglossary regardless of case or spaces
		)
		# a hack that hides the window until we move it to the center of screen
		if os.sep == "\\":  # Windows
			rootWin.attributes("-alpha", 0.0)
		else:  # Linux
			rootWin.withdraw()
		tk.Frame.__init__(self, rootWin)
		UIBase.__init__(self)
		rootWin.title("PyGlossary (Tkinter)")
		rootWin.resizable(True, False)
		# self.progressbarEnable = progressbar
		########
		set_window_icon(rootWin)
		rootWin.bind("<Escape>", lambda _e: rootWin.quit())
		#########
		# Linux: ('clam', 'alt', 'default', 'classic')
		# Windows: ('winnative', 'clam', 'alt', 'default', 'classic', 'vista',
		# 'xpnative')
		style = ttk.Style()
		style.configure("TButton", borderwidth=3)
		# style.theme_use("default")
		# there is no tk.Style()
		########
		self.pack(fill="x")
		# rootWin.bind("<Configure>", self.resized)
		#######################
		defaultFont = tkFont.nametofont("TkDefaultFont")
		if sysName in {"linux", "freebsd"}:
			defaultFont.configure(size=int(defaultFont.cget("size") * 1.4))
		####
		self.bigFont = defaultFont.copy()
		self.bigFont.configure(size=int(defaultFont.cget("size") * 1.6))
		# self.biggerFont = defaultFont.copy()
		# self.biggerFont.configure(size=int(defaultFont.cget("size") * 1.8))
		####
		style.configure(
			"Treeview",
			rowheight=int(defaultFont.metrics("linespace") * 1.5),
		)
		######################
		self.progressbar = progressbar
		self.infoOverride = {}
		self.convertOptions = {}
		self.pathI = ""
		self.pathO = ""
		fcd_dir = join(homeDir, "Desktop")
		if isfile(self.fcd_dir_save_path):
			try:
				with open(self.fcd_dir_save_path, encoding="utf-8") as fp:
					fcd_dir = fp.read().strip("\n")
			except Exception:
				log.exception("")
		self.fcd_dir = fcd_dir
		######################
		notebook = ttk.Notebook(self)
		convertFrame = ttk.Frame(notebook, height=200)
		###################
		row = 0
		ttk.Label(convertFrame, text="Input File: ").grid(
			row=row,
			column=0,
			sticky=tk.W,
			padx=5,
		)
		##
		entry = ttk.Entry(convertFrame)
		entry.grid(
			row=row,
			column=1,
			columnspan=2,
			sticky=tk.W + tk.E,
			padx=0,
		)
		entry.bind_all("<KeyPress>", self.anyEntryChanged)
		self.entryInputConvert = entry
		##
		button = newButton(
			convertFrame,
			text="Browse",
			command=self.browseInputConvert,
			# bg="#f0f000",
			# activebackground="#f6f622",
		)
		button.grid(
			row=row,
			column=3,
			sticky=tk.W + tk.E,
			padx=5,
		)
		######################
		row += 1
		ttk.Label(convertFrame, text="Input Format: ").grid(
			row=row,
			column=0,
			sticky=tk.W,
			padx=5,
		)
		##
		self.formatButtonInputConvert = FormatButton(
			rootWin,
			master=convertFrame,
			descList=readDesc,
			dialogTitle="Select Input Format",
			onChange=self.inputFormatChanged,
		)
		self.formatButtonInputConvert.grid(
			row=row,
			column=1,
			columnspan=2,
			sticky=tk.W,
			padx=0,
		)
		##
		self.readOptions: dict[str, Any] = {}
		self.writeOptions: dict[str, Any] = {}
		######################
		row += 1
		ttk.Label(convertFrame).grid(
			row=row,
			column=0,
			sticky=tk.W,
		)
		######################
		row += 1
		ttk.Label(convertFrame, text="Output File: ").grid(
			row=row,
			column=0,
			sticky=tk.W,
			padx=5,
		)
		##
		entry = ttk.Entry(convertFrame)
		entry.grid(
			row=row,
			column=1,
			columnspan=2,
			sticky=tk.W + tk.E,
			padx=0,
		)
		entry.bind_all("<KeyPress>", self.anyEntryChanged)
		self.entryOutputConvert = entry
		##
		newButton(
			convertFrame,
			text="Browse",
			command=self.browseOutputConvert,
			# bg="#f0f000",
			# activebackground="#f6f622",
		).grid(
			row=row,
			column=3,
			sticky=tk.W + tk.E,
			padx=5,
		)
		###################
		row += 1
		ttk.Label(convertFrame, text="Output Format: ").grid(
			row=row,
			column=0,
			sticky=tk.W,
			padx=5,
		)
		##
		self.formatButtonOutputConvert = FormatButton(
			rootWin,
			master=convertFrame,
			descList=writeDesc,
			dialogTitle="Select Output Format",
			onChange=self.outputFormatChanged,
		)
		self.formatButtonOutputConvert.grid(
			row=row,
			column=1,
			columnspan=2,
			sticky=tk.W,
			padx=0,
		)
		###################
		row += 1
		optionsMenu = tk.Menu(tearoff=False)
		optionsMenu.add_command(
			label="Read Options",
			command=self.readOptionsClicked,
			font=defaultFont,
		)
		optionsMenu.add_command(
			label="Write Options",
			command=self.writeOptionsClicked,
			font=defaultFont,
		)
		optionsMenu.add_command(
			label="General Options",
			command=self.generalOptionsClicked,
			font=defaultFont,
		)
		optionsMenu.add_command(
			label="Info / Metadata",
			command=self.infoOptionClicked,
			font=defaultFont,
		)
		optionsButton = ttk.Menubutton(
			convertFrame,
			text="Options    ",
			menu=optionsMenu,
		)
		optionsButton.grid(
			row=row,
			column=1,
			columnspan=1,
			sticky=tk.E + tk.S,
			padx=5,
			pady=5,
		)
		newButton(
			convertFrame,
			text="Convert",
			command=self.convert,
			# background="#00e000",
			# activebackground="#22f022",
			# borderwidth=7,
			# font=self.biggerFont,
			# padx=5,
			# pady=5,
		).grid(
			row=row,
			column=2,
			columnspan=3,
			sticky=tk.W + tk.E + tk.S,
			padx=5,
			pady=5,
		)
		# print(f"row number for Convert button: {row}")
		#################
		row += 1
		console = tk.Text(
			convertFrame,
			height=15,
			background="#000",
			foreground="#fff",
		)
		console.bind("<Key>", self.consoleKeyPress)
		# self.consoleH = 15
		# sbar = Tix.Scrollbar(
		# 	convertFrame,
		# 	orien=Tix.VERTICAL,
		# 	command=console.yview
		# )
		# sbar.grid (row=row, column=1)
		# console["yscrollcommand"] = sbar.set
		console.grid(
			row=row,
			column=0,
			columnspan=4,
			sticky=tk.W + tk.E,
			padx=5,
			pady=0,
		)
		log.addHandler(
			TkTextLogHandler(console),
		)
		console.insert("end", "Console:\n")
		####
		self.console = console
		##################
		aboutFrame = ttk.Frame(notebook)

		versionFrame = ttk.Frame(aboutFrame, borderwidth=5)
		newLabelWithImage(versionFrame, file=logo).pack(
			side="left", fill="both", expand=False
		)
		ttk.Label(versionFrame, text=f"PyGlossary\nVersion {getVersion()}").pack(
			side="left", fill="both", expand=False
		)
		versionFrame.pack(side="top", fill="x")
		##

		aboutNotebook = VerticalNotebook(aboutFrame, font=self.bigFont)

		aboutAboutFrame = ttk.Frame()
		newReadOnlyText(
			aboutAboutFrame,
			text=f"{aboutText}\nHome page: {homePage}",
			font=("DejaVu Sans", 11, ""),
		).pack(fill="both", expand=True)
		aboutAboutFrame.pack(side="top", fill="x")
		aboutNotebook.add(aboutAboutFrame, "About")

		authorsFrame = ttk.Frame()
		authorsText = "\n".join(authors).replace("\t", "    ")
		newReadOnlyText(
			authorsFrame,
			text=authorsText,
			font=("DejaVu Sans", 11, ""),
		).pack(fill="both", expand=True)
		aboutNotebook.add(authorsFrame, "Authors")

		licenseFrame = ttk.Frame()
		newReadOnlyText(
			licenseFrame,
			text=licenseText,
			font=("DejaVu Sans", 11, ""),
		).pack(fill="both", expand=True)
		aboutNotebook.add(licenseFrame, "License")

		aboutNotebook.pack(fill="both", expand=True)
		# aboutNotebook.show_tab_index(0)

		statusBarFrame = self.statusBarFrame = ttk.Frame(convertFrame)
		statusBarFrame.grid(
			row=row + 1,
			column=0,
			columnspan=4,
			sticky=tk.W + tk.E,
			padx=5,
			pady=0,
		)
		clearB = newButton(
			statusBarFrame,
			text="Clear",
			command=self.console_clear,
			# how to set borderwidth using style?
			# bg="black",
			# fg="#ffff00",
			# activebackground="#333333",
			# activeforeground="#ffff00",
			# borderwidth=3,
			# height=2,
		)
		clearB.pack(side="left")
		####
		ttk.Label(statusBarFrame, text="Verbosity").pack(side="left")
		##
		comboVar = tk.StringVar()
		verbosity = log.getVerbosity()
		combo = ttk.OptionMenu(
			statusBarFrame,
			comboVar,
			f"{verbosity} - {log.levelNamesCap[verbosity]}",
			"0 - Critical",
			"1 - Error",
			"2 - Warning",
			"3 - Info",
			"4 - Debug",
			"5 - Trace",
			"6 - All",
		)

		comboVar.trace_add("write", self.verbosityChanged)
		combo.pack(side="left")
		self.verbosityCombo = comboVar

		notebook.add(convertFrame, text="Convert", underline=-1)
		notebook.add(aboutFrame, text="About", underline=-1)

		# convertFrame.pack(fill="x")
		# convertFrame.grid(sticky=tk.W + tk.E + tk.N + tk.S)

		######################
		for column, weight in enumerate([1, 30, 20, 1]):
			tk.Grid.columnconfigure(convertFrame, column, weight=weight)
		for row, weight in enumerate([50, 50, 1, 50, 50, 1, 50]):
			tk.Grid.rowconfigure(convertFrame, row, weight=weight)
		# _________________________________________________________________ #

		notebook.pack(fill="both", expand=True)

	def openDialog(self, dialog: tk.Toplevel) -> None:
		# x, y, w, h = decodeGeometry(dialog.geometry())
		# w and h are rough estimated width and height of `dialog`
		px, py, pw, ph = decodeGeometry(self.winfo_toplevel().geometry())
		dialog.update_idletasks()
		width = dialog.winfo_width() + 400
		height = dialog.winfo_height()
		dialog.geometry(
			f"{width}x{height}"
			+ encodeLocation(
				px + pw // 2 - width // 2,
				py + ph // 2 - height // 2,
			),
		)
		dialog.focus()

	def readOptionsClicked(self) -> None:
		formatDesc = self.formatButtonInputConvert.get()
		if not formatDesc:
			return

		def okFunc(values: dict[str, Any]) -> None:
			self.readOptions = values

		dialog = FormatOptionsDialog(
			formatDesc,
			"Read",
			self.readOptions,
			okFunc,
			master=self,
		)
		self.openDialog(dialog)

	def writeOptionsClicked(self) -> None:
		formatDesc = self.formatButtonOutputConvert.get()
		if not formatDesc:
			return

		def okFunc(values: dict[str, Any]) -> None:
			self.writeOptions = values

		dialog = FormatOptionsDialog(
			formatDesc=formatDesc,
			kind="Write",
			values=self.writeOptions,
			okFunc=okFunc,
			master=self,
		)
		self.openDialog(dialog)

	def generalOptionsClicked(self) -> None:
		dialog = GeneralOptionsDialog(self, master=self.winfo_toplevel())
		self.openDialog(dialog)

	def infoOptionClicked(self) -> None:
		def okFunc(info: dict[str, Any]) -> None:
			self.infoOverride = info

		dialog = PreConvertInfoDialog(
			info=self.infoOverride,
			okFunc=okFunc,
			master=self,
		)
		dialog.focus()

	def textSelectAll(self, tktext: tk.Text) -> None:
		tktext.tag_add(tk.SEL, "1.0", tk.END)
		tktext.mark_set(tk.INSERT, "1.0")
		tktext.see(tk.INSERT)

	def consoleKeyPress(self, event: tk.Event) -> str | None:
		# print(e.state, e.keysym)
		if event.state > 0:
			if event.keysym == "c":
				return None
			if event.keysym == "a":
				self.textSelectAll(self.console)
				return "break"
		if event.keysym == "Escape":
			return None
		return "break"

	def verbosityChanged(self, _index: str, _value: str, _op: str) -> None:
		log.setVerbosity(
			int(self.verbosityCombo.get()[0]),
		)

	# def resized(self, event):
	# 	self.rootWin.winfo_height() - self.winfo_height()
	# 	log.debug(dh, self.consoleH)
	# 	if dh > 20:
	# 		self.consoleH += 1
	# 		self.console["height"] = self.consoleH
	# 		self.console["width"] = int(self.console["width"]) + 1
	# 		self.console.grid()
	# 	for x in dir(self):
	# 		if "info" in x:
	# 			log.debug(x)

	def inputFormatChangedAuto(self) -> None:
		self.inputFormatChanged(self.formatButtonInputConvert.get())

	def outputFormatChangedAuto(self) -> None:
		self.outputFormatChanged(self.formatButtonOutputConvert.get())

	def inputFormatChanged(self, formatDesc: str) -> None:
		if not formatDesc:
			return
		self.readOptions.clear()  # reset the options, DO NOT re-assign

	def outputFormatChanged(self, formatDesc: str) -> None:
		if not formatDesc:
			return

		formatName = pluginByDesc[formatDesc].name
		plugin = Glossary.plugins.get(formatName)
		if not plugin:
			log.error(f"plugin {formatName} not found")
			return

		self.writeOptions.clear()  # reset the options, DO NOT re-assign

		pathI = self.entryInputConvert.get()
		if (
			pathI
			and not self.entryOutputConvert.get()
			and self.formatButtonInputConvert.get()
			and plugin.extensionCreate
		):
			pathNoExt, _ext = splitext(pathI)
			self.entryOutputConvert.insert(
				0,
				pathNoExt + plugin.extensionCreate,
			)

	def anyEntryChanged(self, _event: tk.Event | None = None) -> None:
		self.inputEntryChanged()
		self.outputEntryChanged()

	def inputEntryChanged(self, _event: tk.Event | None = None) -> None:
		# char = event.keysym
		pathI = self.entryInputConvert.get()
		if self.pathI == pathI:
			return

		if pathI.startswith("file://"):
			pathI = urlToPath(pathI)
			self.entryInputConvert.delete(0, "end")
			self.entryInputConvert.insert(0, pathI)
		if self.config["ui_autoSetFormat"]:
			formatDesc = self.formatButtonInputConvert.get()
			if not formatDesc:
				try:
					inputArgs = Glossary.detectInputFormat(pathI)
				except Error:
					pass
				else:
					plugin = Glossary.plugins.get(inputArgs.formatName)
					if plugin:
						self.formatButtonInputConvert.setValue(plugin.description)
						self.inputFormatChangedAuto()
		self.pathI = pathI

	def outputEntryChanged(self, _event: tk.Event | None = None) -> None:
		pathO = self.entryOutputConvert.get()
		if self.pathO == pathO:
			return

		if pathO.startswith("file://"):
			pathO = urlToPath(pathO)
			self.entryOutputConvert.delete(0, "end")
			self.entryOutputConvert.insert(0, pathO)
		if self.config["ui_autoSetFormat"]:
			formatDesc = self.formatButtonOutputConvert.get()
			if not formatDesc:
				try:
					outputArgs = Glossary.detectOutputFormat(
						filename=pathO,
						inputFilename=self.entryInputConvert.get(),
					)
				except Error:
					pass
				else:
					self.formatButtonOutputConvert.setValue(
						Glossary.plugins[outputArgs.formatName].description,
					)
					self.outputFormatChangedAuto()
		self.pathO = pathO

	def save_fcd_dir(self) -> None:
		if not self.fcd_dir:
			return
		with open(self.fcd_dir_save_path, mode="w", encoding="utf-8") as fp:
			fp.write(self.fcd_dir)

	def browseInputConvert(self) -> None:
		path = filedialog.askopenfilename(initialdir=self.fcd_dir)
		if path:
			self.entryInputConvert.delete(0, "end")
			self.entryInputConvert.insert(0, path)
			self.inputEntryChanged()
			self.fcd_dir = os.path.dirname(path)
			self.save_fcd_dir()

	def browseOutputConvert(self) -> None:
		path = filedialog.asksaveasfilename()
		if path:
			self.entryOutputConvert.delete(0, "end")
			self.entryOutputConvert.insert(0, path)
			self.outputEntryChanged()
			self.fcd_dir = os.path.dirname(path)
			self.save_fcd_dir()

	def convert(self) -> None:
		inPath = self.entryInputConvert.get()
		if not inPath:
			log.critical("Input file path is empty!")
			return
		inFormatDesc = self.formatButtonInputConvert.get()
		# if not inFormatDesc:
		# 	log.critical("Input format is empty!");return
		inFormat = pluginByDesc[inFormatDesc].name if inFormatDesc else ""

		outPath = self.entryOutputConvert.get()
		if not outPath:
			log.critical("Output file path is empty!")
			return
		outFormatDesc = self.formatButtonOutputConvert.get()
		if not outFormatDesc:
			log.critical("Output format is empty!")
			return
		outFormat = pluginByDesc[outFormatDesc].name

		log.debug(f"config: {self.config}")

		glos = Glossary(ui=self)
		glos.config = self.config
		glos.progressbar = self.progressbar

		for attr, value in self._glossarySetAttrs.items():
			setattr(glos, attr, value)

		if self.infoOverride:
			log.info(f"infoOverride = {self.infoOverride}")

		if self.convertOptions:
			log.info(f"convertOptions: {self.convertOptions}")

		try:
			glos.convert(
				ConvertArgs(
					inPath,
					inputFormat=inFormat,
					outputFilename=outPath,
					outputFormat=outFormat,
					readOptions=self.readOptions,
					writeOptions=self.writeOptions,
					infoOverride=self.infoOverride or None,
					**self.convertOptions,
				),
			)
		except Error as e:
			log.critical(str(e))
			glos.cleanup()
			return

		# self.status("Convert finished")

	def run(  # noqa: PLR0913
		self,
		inputFilename: str = "",
		outputFilename: str = "",
		inputFormat: str = "",
		outputFormat: str = "",
		reverse: bool = False,
		config: ConfigType | None = None,
		readOptions: dict[str, Any] | None = None,
		writeOptions: dict[str, Any] | None = None,
		convertOptions: dict[str, Any] | None = None,
		glossarySetAttrs: dict[str, Any] | None = None,
	) -> None:
		config = config or {}
		self.config = config

		if inputFilename:
			self.entryInputConvert.insert(0, abspath2(inputFilename))
			self.inputEntryChanged()
		if outputFilename:
			self.entryOutputConvert.insert(0, abspath2(outputFilename))
			self.outputEntryChanged()

		if inputFormat and inputFormat not in Glossary.readFormats:
			log.error(f"invalid {inputFormat=}")
			inputFormat = ""

		if outputFormat and outputFormat not in Glossary.writeFormats:
			log.error(f"invalid {outputFormat=}")
			outputFormat = ""

		if inputFormat:
			self.formatButtonInputConvert.setValue(
				Glossary.plugins[inputFormat].description,
			)
			self.inputFormatChangedAuto()
		if outputFormat:
			self.formatButtonOutputConvert.setValue(
				Glossary.plugins[outputFormat].description,
			)
			self.outputFormatChangedAuto()

		if reverse:
			log.error("Tkinter interface does not support Reverse feature")

		pbar = ProgressBar(
			self.statusBarFrame,
			min_=0,
			max_=100,
			width=700,
			height=28,
			appearance="sunken",
			fillColor=config.get("tk.progressbar.color.fill", "blue"),
			background=config.get("tk.progressbar.color.background", "gray"),
			labelColor=config.get("tk.progressbar.color.text", "yellow"),
			labelFont=config.get("tk.progressbar.font", "Sans"),
		)
		pbar.pack(side="left", fill="x", expand=True, padx=10)
		self.pbar = pbar
		pbar.pack(fill="x")
		self.progressTitle = ""
		# _________________________________________________________________ #

		centerWindow(self.rootWin)
		# show the window
		if os.sep == "\\":  # Windows
			self.rootWin.attributes("-alpha", 1.0)
		else:  # Linux
			self.rootWin.deiconify()

		# must be before setting self.readOptions and self.writeOptions
		self.anyEntryChanged()

		if readOptions:
			self.readOptions = readOptions

		if writeOptions:
			self.writeOptions = writeOptions

		self.convertOptions: dict[str, Any] = convertOptions or {}
		if convertOptions:
			log.info(f"Using {convertOptions=}")

		self._glossarySetAttrs = glossarySetAttrs or {}

		# inputFilename and readOptions are for DB Editor
		# which is not implemented
		self.mainloop()

	def progressInit(self, title: str) -> None:
		self.progressTitle = title

	def progress(self, ratio: float, text: str = "") -> None:
		if not text:
			text = "%" + str(int(ratio * 100))
		text += " - " + self.progressTitle
		self.pbar.updateProgress(ratio * 100, None, text)
		# self.pbar.value = ratio * 100
		# self.pbar.update()
		self.rootWin.update()

	def console_clear(self, _event: tk.Event | None = None) -> None:
		self.console.delete("1.0", "end")
		self.console.insert("end", "Console:\n")

	# def reverseBrowseInput(self):
	# 	pass

	# def reverseBrowseOutput(self):
	# 	pass

	# def reverseLoad(self):
	# 	pass



from __future__ import annotations

import sys
import unittest
from io import BytesIO
from os.path import abspath, dirname
from typing import TYPE_CHECKING, cast

import lxml.html
from lxml import etree as ET

if TYPE_CHECKING:
	from pyglossary.lxml_types import T_htmlfile

rootDir = dirname(dirname(abspath(__file__)))
sys.path.insert(0, rootDir)

from pyglossary.plugins.edict2.conv import (
	render_definition_no_links,
	render_definition_with_links,
	render_syllables,
)
from pyglossary.plugins.edict2.pinyin import convert


class PinyinTest(unittest.TestCase):
	@staticmethod
	def render(pinyin: str) -> str:
		pinyin_list, tones = zip(*map(convert, pinyin.split()), strict=False)
		f = BytesIO()
		with ET.htmlfile(f, encoding="utf-8") as hf_:  # noqa: PLR1702
			hf = cast("T_htmlfile", hf_)
			render_syllables(hf, pinyin_list, tones)

		result = f.getvalue().decode("utf-8")

		# remove html tags
		return lxml.html.fromstring(result).text_content()

	def test_pinyin_names(self):
		self.assertEqual(self.render("Chang2 an1"), "Cháng'ān")
		self.assertEqual(self.render("Sun1 Zhong1 shan1"), "Sūn Zhōngshān")
		self.assertEqual(self.render("A1 Q Zheng4 zhuan4"), "ĀQ Zhèngzhuàn")
		self.assertEqual(self.render("Q tan2"), "Qtán")
		self.assertEqual(self.render("E1 mi2 tuo2 Fo2"), "Ēmítuó Fó")

	def test_pinyin_apostrophes(self):
		self.assertEqual(self.render("qi3 e2"), "qǐ'é")
		self.assertEqual(self.render("e2"), "é")
		self.assertEqual(self.render("Li3 An1"), "Lǐ Ān")

	def test_pinyin_umlaut_u(self):
		self.assertEqual(self.render("gui1 nü"), "guīnü")
		self.assertEqual(self.render("hu1 lüe4"), "hūlüè")


class DefinitionLinkTest(unittest.TestCase):
	@staticmethod
	def render_no_links(definition: str) -> str:
		f = BytesIO()
		with ET.htmlfile(f, encoding="utf-8") as hf_:
			hf = cast("T_htmlfile", hf_)
			render_definition_no_links(hf, definition=definition)

		return f.getvalue().decode("utf-8")

	@staticmethod
	def render_with_links_simplified(definition: str) -> str:
		f = BytesIO()
		with ET.htmlfile(f, encoding="utf-8") as hf_:
			hf = cast("T_htmlfile", hf_)
			render_definition_with_links(
				hf,
				definition=definition,
				traditional_title=False,
			)

		return f.getvalue().decode("utf-8")

	@staticmethod
	def render_with_links_traditional(definition: str) -> str:
		f = BytesIO()
		with ET.htmlfile(f, encoding="utf-8") as hf_:
			hf = cast("T_htmlfile", hf_)
			render_definition_with_links(
				hf,
				definition=definition,
				traditional_title=True,
			)

		return f.getvalue().decode("utf-8")

	def test_links_created_simplified(self):
		self.assertEqual(
			self.render_with_links_simplified("CL:個|个[ge4],隻|只[zhi1]/"),
			'<li>CL:<a href="bword://个">個|个[ge4]</a>,<a href="bword://只">隻|只[zhi1]</a>/</li>',
		)
		self.assertEqual(
			self.render_with_links_simplified(
				"provincial 解元[jie4 yuan2], metropolitan 會元|会元[hui4 yuan2] and palace 狀元|状元[zhuang4 yuan2]"
			),
			'<li>provincial <a href="bword://解元">解元[jie4 yuan2]</a>, metropolitan <a href="bword://会元">會元|会元[hui4 yuan2]</a> and palace <a href="bword://状元">狀元|状元[zhuang4 yuan2]</a></li>',
		)

	def test_links_created_traditional(self):
		self.assertEqual(
			self.render_with_links_traditional("CL:個|个[ge4],隻|只[zhi1]/"),
			'<li>CL:<a href="bword://個">個|个[ge4]</a>,<a href="bword://隻">隻|只[zhi1]</a>/</li>',
		)
		self.assertEqual(
			self.render_with_links_traditional(
				"provincial 解元[jie4 yuan2], metropolitan 會元|会元[hui4 yuan2] and palace 狀元|状元[zhuang4 yuan2]"
			),
			'<li>provincial <a href="bword://解元">解元[jie4 yuan2]</a>, metropolitan <a href="bword://會元">會元|会元[hui4 yuan2]</a> and palace <a href="bword://狀元">狀元|状元[zhuang4 yuan2]</a></li>',
		)

	def test_no_links(self):
		for defn in (
			"CL:個|个[ge4],隻|只[zhi1]/",
			"provincial 解元[jie4 yuan2], metropolitan 會元|会元[hui4 yuan2] and palace 狀元|状元[zhuang4 yuan2]",
		):
			self.assertEqual(self.render_no_links(defn), f"<li>{defn}</li>")


if __name__ == "__main__":
	unittest.main()



import os
import struct
import sys
import unittest
from os.path import abspath, dirname

rootDir = dirname(dirname(abspath(__file__)))
sys.path.insert(0, rootDir)

from pyglossary.text_utils import (
	crc32hex,
	escapeNTB,
	fixUtf8,
	joinByBar,
	replacePostSpaceChar,
	splitByBar,
	splitByBarUnescapeNTB,
	uint32FromBytes,
	uint32ToBytes,
	uintFromBytes,
	unescapeBar,
	unescapeNTB,
	urlToPath,
)


class TestTextUtils(unittest.TestCase):
	def test_fixUtf8(self):
		f = fixUtf8
		# Since entries already keep words and defi as string, fixUtf8 does not
		# do much. It just removes zero bytes between valid characters
		# (and not within characters)
		# If there were encoding errors in input file, Reader class would
		# most likely fail to read and raise exception.
		# This feature was useful in Python 2.x, but not much anymore!
		self.assertEqual(f("\x00س\x00لام"), "سلام")

	def test_unescapeNTB(self):
		self.assertEqual("a", unescapeNTB("a", bar=False))
		self.assertEqual("a\t", unescapeNTB("a\\t", bar=False))
		self.assertEqual("a\n", unescapeNTB("a\\n", bar=False))
		self.assertEqual("\ta", unescapeNTB("\\ta", bar=False))
		self.assertEqual("\na", unescapeNTB("\\na", bar=False))
		self.assertEqual("a\tb\n", unescapeNTB("a\\tb\\n", bar=False))
		self.assertEqual("a\\b", unescapeNTB("a\\\\b", bar=False))
		self.assertEqual("a\\\tb", unescapeNTB("a\\\\\\tb", bar=False))
		self.assertEqual("a|b\tc", unescapeNTB("a|b\\tc", bar=False))
		self.assertEqual("a\\|b\tc", unescapeNTB("a\\|b\\tc", bar=False))
		self.assertEqual("a\\|b\tc", unescapeNTB("a\\\\|b\\tc", bar=False))
		self.assertEqual("|", unescapeNTB("\\|", bar=True))
		self.assertEqual("a|b", unescapeNTB("a\\|b", bar=True))
		self.assertEqual("a|b\tc", unescapeNTB("a\\|b\\tc", bar=True))

	def test_escapeNTB(self):
		self.assertEqual(escapeNTB("a", bar=False), "a")
		self.assertEqual(escapeNTB("a\t", bar=False), "a\\t")
		self.assertEqual(escapeNTB("a\n", bar=False), "a\\n")
		self.assertEqual(escapeNTB("\ta", bar=False), "\\ta")
		self.assertEqual(escapeNTB("\na", bar=False), "\\na")
		self.assertEqual(escapeNTB("a\tb\n", bar=False), "a\\tb\\n")
		self.assertEqual(escapeNTB("a\\b", bar=False), "a\\\\b")
		self.assertEqual(escapeNTB("a\\\tb", bar=False), "a\\\\\\tb")
		self.assertEqual(escapeNTB("a|b\tc", bar=False), "a|b\\tc")
		self.assertEqual(escapeNTB("a\\|b\tc", bar=False), "a\\\\|b\\tc")
		self.assertEqual(escapeNTB("|", bar=True), "\\|")
		self.assertEqual(escapeNTB("a|b", bar=True), "a\\|b")
		self.assertEqual(escapeNTB("a|b\tc", bar=True), "a\\|b\\tc")

	def test_splitByBarUnescapeNTB(self):
		f = splitByBarUnescapeNTB
		self.assertEqual(f(""), [""])
		self.assertEqual(f("|"), ["", ""])
		self.assertEqual(f("a"), ["a"])
		self.assertEqual(f("a|"), ["a", ""])
		self.assertEqual(f("|a"), ["", "a"])
		self.assertEqual(f("a|b"), ["a", "b"])
		self.assertEqual(f("a\\|b|c"), ["a|b", "c"])
		self.assertEqual(f("a\\\\1|b|c"), ["a\\1", "b", "c"])
		# self.assertEqual(f("a\\\\|b|c"), ["a\\", "b", "c"])  # FIXME
		self.assertEqual(f("a\\\\1|b\\n|c\\t"), ["a\\1", "b\n", "c\t"])

	def test_unescapeBar(self):
		f = unescapeBar
		self.assertEqual("", f(""))
		self.assertEqual("|", f("\\|"))
		self.assertEqual("a|b", f("a\\|b"))
		self.assertEqual("a|b\tc", f("a\\|b\tc"))
		self.assertEqual("a|b\\t\\nc", f("a\\|b\\t\\nc"))
		self.assertEqual("\\", f("\\\\"))
		self.assertEqual("\\|", f("\\\\\\|"))

	def test_splitByBar(self):
		f = splitByBar
		self.assertEqual(f(""), [""])
		self.assertEqual(f("|"), ["", ""])
		self.assertEqual(f("a"), ["a"])
		self.assertEqual(f("a|"), ["a", ""])
		self.assertEqual(f("|a"), ["", "a"])
		self.assertEqual(f("a|b"), ["a", "b"])
		self.assertEqual(f("a\\|b"), ["a|b"])
		self.assertEqual(f("a\\|b|c"), ["a|b", "c"])
		self.assertEqual(f("a\\\\1|b|c"), ["a\\1", "b", "c"])
		# self.assertEqual(f("a\\\\|b|c"), ["a\\", "b", "c"])  # FIXME

	def test_joinByBar(self):
		f = joinByBar
		self.assertEqual("", f([""]))
		self.assertEqual("|", f(["", ""]))
		self.assertEqual("a", f(["a"]))
		self.assertEqual("a|", f(["a", ""]))
		self.assertEqual("|a", f(["", "a"]))
		self.assertEqual("a|b", f(["a", "b"]))
		self.assertEqual("a\\|b", f(["a|b"]))
		self.assertEqual("a\\|b|c", f(["a|b", "c"]))
		self.assertEqual("a\\\\1|b|c", f(["a\\1", "b", "c"]))

	def test_uint32ToBytes(self):
		f = uint32ToBytes

		outOfRangeError = "'I' format requires 0 <= number <= 4294967295"
		if os.sep == "\\":
			outOfRangeError = "argument out of range"

		self.assertEqual(f(0), bytes([0, 0, 0, 0]))
		self.assertEqual(f(0x3E8), bytes([0, 0, 0x03, 0xE8]))
		self.assertEqual(f(0x186A0), bytes([0, 1, 0x86, 0xA0]))
		self.assertEqual(f(0x3B9ACA00), bytes([0x3B, 0x9A, 0xCA, 0x00]))
		self.assertEqual(f(0xFFFFFFFF), bytes([0xFF, 0xFF, 0xFF, 0xFF]))

		with self.assertRaises(struct.error) as ctx:
			f(0xFFFFFFFF + 1)
		self.assertEqual(
			str(ctx.exception),
			outOfRangeError,
		)

		with self.assertRaises(struct.error) as ctx:
			f(10000000000)
		self.assertEqual(
			str(ctx.exception),
			outOfRangeError,
		)

		with self.assertRaises(struct.error) as ctx:
			f(-1)
		if sys.version_info >= (3, 12):
			self.assertEqual(
				str(ctx.exception),
				"'I' format requires 0 <= number <= 4294967295",
			)
		else:
			self.assertEqual(str(ctx.exception), "argument out of range")

	def test_uint32FromBytes(self):
		f = uint32FromBytes
		self.assertEqual(0, f(bytes([0, 0, 0, 0])))
		self.assertEqual(0x3E8, f(bytes([0, 0, 0x03, 0xE8])))
		self.assertEqual(0x186A0, f(bytes([0, 1, 0x86, 0xA0])))
		self.assertEqual(0x3B9ACA00, f(bytes([0x3B, 0x9A, 0xCA, 0x00])))
		self.assertEqual(0xFFFFFFFF, f(bytes([0xFF, 0xFF, 0xFF, 0xFF])))

		with self.assertRaises(struct.error) as ctx:
			f(bytes([0x01, 0xFF, 0xFF, 0xFF, 0xFF]))
		self.assertEqual(str(ctx.exception), "unpack requires a buffer of 4 bytes")

	def test_uintFromBytes(self):
		f = uintFromBytes
		self.assertEqual(0, f(bytes([0, 0, 0, 0])))
		self.assertEqual(0x3E8, f(bytes([0, 0, 0x03, 0xE8])))
		self.assertEqual(0x186A0, f(bytes([0, 1, 0x86, 0xA0])))
		self.assertEqual(0x3B9ACA00, f(bytes([0x3B, 0x9A, 0xCA, 0x00])))
		self.assertEqual(0xFFFFFFFF, f(bytes([0xFF, 0xFF, 0xFF, 0xFF])))
		self.assertEqual(
			0xFFABCDEF5542,
			f(bytes([0xFF, 0xAB, 0xCD, 0xEF, 0x55, 0x42])),
		)

	def test_crc32hex(self):
		f = crc32hex
		self.assertEqual(f(b""), "00000000")
		self.assertEqual(f(b"\x00"), "d202ef8d")
		self.assertEqual(f(b"\x00\x00"), "41d912ff")
		self.assertEqual(
			f(bytes.fromhex("73c3bbc38b7459360ac3a9c2b3c2a2")),
			"bbfb1610",
		)

	def test_urlToPath(self):
		f = urlToPath
		self.assertEqual(
			f("https://github.com/ilius/pyglossary"),
			"https://github.com/ilius/pyglossary",
		)
		self.assertEqual(
			f("file:///home/test/abc.txt"),
			"/home/test/abc.txt",
		)
		self.assertEqual(
			f("file:///home/test/%D8%AA%D8%B3%D8%AA.txt"),
			"/home/test/تست.txt",
		)

	def test_replacePostSpaceChar(self):
		f = replacePostSpaceChar
		self.assertEqual(
			f("First sentence .Second sentence.", "."),
			"First sentence. Second sentence.",
		)
		self.assertEqual(
			f("First ,second.", ","),
			"First, second.",
		)


if __name__ == "__main__":
	unittest.main()



import sys
import tempfile
import unittest
from os.path import abspath, dirname, join

rootDir = dirname(dirname(abspath(__file__)))
sys.path.insert(0, rootDir)

from glossary_v2_test import TestGlossaryBase

from pyglossary.glossary_v2 import ConvertArgs, Glossary


class TestGlossaryDSL(TestGlossaryBase):
	def __init__(self, *args, **kwargs):
		TestGlossaryBase.__init__(self, *args, **kwargs)

		self.dataFileCRC32.update(
			{
				"dsl/100-RussianAmericanEnglish-ru-en.dsl": "c24491e0",
				"dsl/100-RussianAmericanEnglish-ru-en-v3.txt": "43b6d58e",
				"dsl/001-empty-lines-br.dsl": "6f2fca1a",
				"dsl/001-empty-lines-br.txt": "74e578ff",
				"dsl/002-m-tag_multiline-paragraph.dsl": "334079e3",
				"dsl/002-m-tag_multiline-paragraph-v2.txt": "d5001afd",
				"dsl/003-ref-target-c.dsl": "9c1396c4",
				"dsl/003-ref-target-c.txt": "ab41cedf",
				"dsl/006-include/included.dsl": "10ef1018",
				"dsl/006-include/main.dsl": "8325d538",
				"dsl/006-include/main.txt": "d7ed9f2b",
			},
		)

	def convert_string_dsl_txt(
		self,
		dsl: str,
		txtExpected: str,
		removeInfo: bool = True,
		**convertArgs,
	):
		prefix = join(self.tempDir, "")
		dsl_fname = tempfile.mktemp(suffix=".dsl", prefix=prefix)
		txt_fname = tempfile.mktemp(suffix=".txt", prefix=prefix)
		with open(dsl_fname, "w", encoding="utf-8") as file:
			file.write(dsl)

		glos = self.glos = Glossary()
		# glos.config = config
		res = glos.convert(
			ConvertArgs(
				inputFilename=dsl_fname,
				outputFilename=txt_fname,
				**convertArgs,
			),
		)
		self.assertEqual(txt_fname, res)

		with open(txt_fname, encoding="utf-8") as file:
			txtActual = file.read()

		if removeInfo:
			txtActual = "\n".join(
				[
					line
					for line in txtActual.split("\n")
					if line and not line.startswith("#")
				],
			)

		self.assertEqual(txtExpected, txtActual)

	def convert_dsl_txt(self, fname, fname2, **convertArgs):
		self.convert(
			f"dsl/{fname}.dsl",
			f"{fname}-2.txt",
			compareText=f"dsl/{fname2}.txt",
			**convertArgs,
		)

	def test_include(self):
		dirName = "dsl/006-include"
		fname = "main"
		files = ["included.dsl", "main.dsl"]

		inputDirPath = self.downloadDir(dirName, files)
		inputFilePath = join(inputDirPath, f"{fname}.dsl")
		outputFilePath = self.newTempFilePath(f"{fname}-2.txt")
		expectedOutputFilePath = self.downloadFile(join(dirName, f"{fname}.txt"))
		self.glos = Glossary()
		result = self.glos.convert(
			ConvertArgs(
				inputFilename=inputFilePath,
				outputFilename=outputFilePath,
			)
		)
		self.assertIsNotNone(result)
		self.assertEqual(result, outputFilePath)

		self.compareTextFiles(
			outputFilePath,
			expectedOutputFilePath,
		)

	def test_russianAmericanEnglish(self):
		self.convert_dsl_txt(
			"100-RussianAmericanEnglish-ru-en",
			"100-RussianAmericanEnglish-ru-en-v3",
		)

	def test_empty_lines_br(self):
		self.convert_dsl_txt(
			"001-empty-lines-br",
			"001-empty-lines-br",
		)

	def test_m_tag_multiline_paragraph(self):
		self.convert_dsl_txt(
			"002-m-tag_multiline-paragraph",
			"002-m-tag_multiline-paragraph-v2",
		)

	def test_ref_target_c(self):
		self.convert_dsl_txt(
			"003-ref-target-c",
			"003-ref-target-c",
		)

	def test_headword_formatting_bashkir_basque(self):
		# from Bashkir -> Basque dict (001-headword-with-formatting.dsl)
		dsl = (
			"{[c slategray]}{to }{[/c]}tell "
			"{[c slategray]}smb{[/c]} how to do "
			"{[c slategray]}smth{[/c]}\n    [m1][trn]"
			"рассказать кому-либо, как что-либо делать[/trn][/m]"
		)
		txt = (
			"tell smb how to do smth\t"
			'<b><font color="slategray">to </font>tell '
			'<font color="slategray">smb</font> how to do '
			'<font color="slategray">smth</font></b><br/>'
			'<p style="padding-left:1em;margin:0">'
			"рассказать кому-либо, как что-либо делать</p>"
		)
		self.convert_string_dsl_txt(dsl, txt)

	def test_headword_formatting_english(self):
		dsl = (
			"{[c slategray]}{to }{[/c]}tell"
			" {[c violet]}smb{[/c]} {[u]}how{[/u]}"
			" to do {[c violet]}smth{[/c]} {[sub]subscript[/sub]}\n"
			"   [m1]1. main meaning[/m]\n"
			"   [m2]a. first submeaning[/m]\n"
			"   [m2]b. second submeaning[/m]\n"
		)
		txt = (
			"tell smb how to do smth\t"
			'<b><font color="slategray">to </font>tell'
			' <font color="violet">smb</font> <u>how</u>'
			' to do <font color="violet">smth</font> <sub>subscript</sub></b><br/>'
			'<p style="padding-left:1em;margin:0">1. main meaning</p>'
			'<p style="padding-left:2em;margin:0">a. first submeaning</p>'
			'<p style="padding-left:2em;margin:0">b. second submeaning</p>'
		)
		self.convert_string_dsl_txt(dsl, txt)

	def test_p_unclosed(self):
		dsl = "headword\n   [m1][p]test\n"
		txt = (
			"headword\t"
			'<p style="padding-left:1em;margin:0">'
			'<i class="p"><font color="green">test\\n</font></i>'
		)
		self.convert_string_dsl_txt(dsl, txt)

	def test_headword_paran(self):
		self.convert_string_dsl_txt(
			"headword with (parenthesis)\n    test",
			"headword with parenthesis|headword with\ttest",
		)

	def test_headword_paran_2(self):
		self.convert_string_dsl_txt(
			"(headword with) parenthesis\n    test",
			"headword with parenthesis|parenthesis\ttest",
		)

	def test_headword_paran_escaped(self):
		self.convert_string_dsl_txt(
			"headword \\(with escaped parenthesis\\)\n    test",
			"headword (with escaped parenthesis)\ttest",
		)

	def test_headword_paran_escaped_2(self):
		self.convert_string_dsl_txt(
			"headword (with escaped right \\) parenthesis)\n    test",
			"headword with escaped right \\\\) parenthesis|headword\ttest",
		)

	def test_headword_curly(self):
		txt = (
			"headword with curly brackets\t"
			"<b>headword with <b>curly brackets</b></b><br/>test"
		)
		self.convert_string_dsl_txt(
			"headword with {[b]}curly brackets{[/b]}\n    test",
			txt,
		)

	def test_headword_curly_escaped(self):
		self.convert_string_dsl_txt(
			"headword with escaped \\{\\}curly brackets\\{\n    test",
			"headword with escaped {}curly brackets{\ttest",
		)

	def test_double_brackets_1(self):
		self.convert_string_dsl_txt(
			"test\n    hello [[world]]",
			"test\thello [world]",
		)

	def test_double_brackets_2(self):
		self.convert_string_dsl_txt(
			"test\n    hello [[",
			"test\thello [",
		)

	def test_double_brackets_3(self):
		self.convert_string_dsl_txt(
			"test\n    hello ]]",
			"test\thello ]",
		)

	def test_ref_double_ltgt(self):
		self.convert_string_dsl_txt(
			"test\n    hello <<world>>",
			'test\thello <a href="bword://world">world</a>',
		)

	def test_ref_double_ltgt_escaped(self):
		self.convert_string_dsl_txt(
			"test\n    hello \\<<world\\>>",
			"test\thello &lt;&lt;world&gt;&gt;",
		)


if __name__ == "__main__":
	unittest.main()



import unittest

from glossary_v2_test import TestGlossaryBase


class TestGlossaryStarDictTextual(TestGlossaryBase):
	def __init__(self, *args, **kwargs):
		TestGlossaryBase.__init__(self, *args, **kwargs)

		self.dataFileCRC32.update(
			{
				"100-en-fa-sdt.xml": "48cb3336",
				"100-en-fa-sdt.xml.txt": "0c9b4025",
				"stardict-xdxf-2.xml": "b3285d5c",
				"stardict-xdxf-2.xml-h.txt": "97b3a22b",
				"stardict-xdxf-2.xml-x.txt": "de63f937",
				"stardict-mixed-types-2.xml": "51d9ceb2",
				"stardict-mixed-types-2.xml.txt": "c896cf68",
			},
		)

	def convert_txt_sdxml(self, fname, fname2, **convertArgs):
		self.convert(
			f"{fname}.txt",
			f"{fname}-2.xml",
			compareText=f"{fname2}.xml",
			outputFormat="StardictTextual",
			**convertArgs,
		)

	def convert_sdxml_txt(self, fname, fname2, **convertArgs):
		self.convert(
			f"{fname}.xml",
			f"{fname}-2.txt",
			compareText=f"{fname2}.txt",
			inputFormat="StardictTextual",
			**convertArgs,
		)

	def test_convert_txt_sdxml_1(self):
		self.convert_txt_sdxml(
			"100-en-fa",
			"100-en-fa-sdt",
		)

	def test_convert_sdxml_txt_1(self):
		self.convert_sdxml_txt(
			"100-en-fa-sdt",
			"100-en-fa-sdt.xml",
		)

	def test_convert_sdxml_txt_2(self):
		self.convert_sdxml_txt(
			"stardict-mixed-types-2",
			"stardict-mixed-types-2.xml",
		)

	def test_convert_sdxml_txt_3(self):
		self.convert_sdxml_txt(
			"stardict-xdxf-2",
			"stardict-xdxf-2.xml-h",
			readOptions={"xdxf_to_html": True},
		)

	def test_convert_sdxml_txt_4(self):
		self.convert_sdxml_txt(
			"stardict-xdxf-2",
			"stardict-xdxf-2.xml-x",
			readOptions={"xdxf_to_html": False},
		)


if __name__ == "__main__":
	unittest.main()



from __future__ import annotations

import hashlib
import json
import logging
import os
import random
import sys
import tempfile
import tracemalloc
import unittest
import zipfile
from os.path import abspath, dirname, isdir, isfile, join
from urllib.request import urlopen

rootDir = dirname(dirname(dirname(abspath(__file__))))
sys.path.insert(0, rootDir)

from typing import TYPE_CHECKING, Any

from pyglossary.core import cacheDir, log, tmpDir
from pyglossary.glossary import Glossary
from pyglossary.os_utils import rmtree
from pyglossary.text_utils import crc32hex

if TYPE_CHECKING:
	from collections.abc import Callable

__all__ = ["TestGlossaryBase", "appTmpDir", "testCacheDir"]


tracemalloc.start()

Glossary.init()

repo = os.getenv(
	"PYGLOSSARY_TEST_REPO",
	"ilius/pyglossary-test/main",
)

dataURL = f"https://raw.githubusercontent.com/{repo}/{{filename}}"

testCacheDir = join(cacheDir, "test")
appTmpDir = join(cacheDir, "tmp")


os.makedirs(testCacheDir, exist_ok=True)
os.chdir(testCacheDir)

os.makedirs(join(tmpDir, "pyglossary"), exist_ok=True)


class TestGlossaryBase(unittest.TestCase):
	def __init__(self, *args, **kwargs):
		unittest.TestCase.__init__(self, *args, **kwargs)
		self.maxDiff = None
		self.dataFileCRC32 = {
			"004-bar.txt": "6775e590",
			"004-bar-sort.txt": "fe861123",
			"006-empty.txt": "07ff224b",
			"006-empty-filtered.txt": "2b3c1c0f",
			"100-en-de-v4.txt": "d420a669",
			"100-en-fa.txt": "f5c53133",
			"100-ja-en.txt": "93542e89",
			"100-en-de-v4-remove_font_b.txt": "a3144e2f",
			"100-en-fa-v2.info": "7c0f646b",
			"300-rand-en-fa.txt": "586617c8",
			"res/stardict.png": "7e1447fa",
			"res/test.json": "41f8cf31",
		}
		os.environ["CALC_FILE_SIZE"] = "1"

	def addDirCRC32(self, dirPath: str, files: dict[str, str]) -> None:
		for fpath, hash_ in files.items():
			self.dataFileCRC32[f"{dirPath}/{fpath}"] = hash_

	# The setUp() and tearDown() methods allow you to define instructions that
	# will be executed before and after each test method.

	def setUp(self):
		self.glos = None
		self.tempDir = tempfile.mkdtemp(dir=join(tmpDir, "pyglossary"))

	def tearDown(self):
		if self.glos is not None:
			self.glos.cleanup()
			self.glos.clear()
		if os.getenv("NO_CLEANUP"):
			return
		for direc in (
			self.tempDir,
			appTmpDir,
		):
			if isdir(direc):
				rmtree(direc)

	def fixDownloadFilename(self, filename):
		return filename.replace("/", "__").replace("\\", "__")

	def downloadFile(self, filename):
		unixFilename = filename.replace("\\", "/")
		crc32 = self.dataFileCRC32[unixFilename]
		fpath = join(testCacheDir, self.fixDownloadFilename(filename))
		if isfile(fpath):
			with open(fpath, mode="rb") as file:
				data = file.read()
			if crc32hex(data) != crc32:
				raise RuntimeError(f"CRC32 check failed for existing file: {fpath}")
			return fpath
		try:
			with urlopen(dataURL.format(filename=unixFilename)) as res:
				data = res.read()
		except Exception as e:
			print(f"{filename=}")
			raise e from None
		actual_crc32 = crc32hex(data)
		if actual_crc32 != crc32:
			raise RuntimeError(
				f"CRC32 check failed for downloaded file: {filename}: {actual_crc32}",
			)
		with open(fpath, mode="wb") as file:
			file.write(data)
		return fpath

	def downloadDir(self, dirName: str, files: list[str]) -> str:
		dirPath = join(testCacheDir, self.fixDownloadFilename(dirName))
		for fileRelPath in files:
			newFilePath = join(dirPath, fileRelPath)
			if isfile(newFilePath):
				# TODO: check crc-32
				continue
			filePath = self.downloadFile(join(dirName, fileRelPath))
			os.makedirs(dirname(newFilePath), exist_ok=True)
			os.rename(filePath, newFilePath)
		return dirPath

	def newTempFilePath(self, filename):
		fpath = join(self.tempDir, filename)
		if isfile(fpath):
			os.remove(fpath)
		return fpath

	def showGlossaryDiff(self, fpath1, fpath2) -> None:
		from pyglossary.ui.tools.diff_glossary import diffGlossary

		diffGlossary(fpath1, fpath2)

	def compareTextFiles(self, fpath1, fpath2, showDiff=False):
		self.assertTrue(isfile(fpath1), f"{fpath1 = }")
		self.assertTrue(isfile(fpath2), f"{fpath2 = }")
		with open(fpath1, encoding="utf-8") as file1:
			text1 = file1.read().rstrip("\n")
		with open(fpath2, encoding="utf-8") as file2:
			text2 = file2.read().rstrip("\n")

		try:
			self.assertEqual(
				len(text1),
				len(text2),
				msg=f"{fpath1} differs from {fpath2} in file size",
			)
			self.assertEqual(
				text1,
				text2,
				msg=f"{fpath1} differs from {fpath2}",
			)
		except AssertionError as e:
			if showDiff:
				self.showGlossaryDiff(fpath1, fpath2)
			raise e from None

	def compareBinaryFiles(self, fpath1, fpath2):
		self.assertTrue(isfile(fpath1), f"File {fpath1} does not exist")
		self.assertTrue(isfile(fpath2), f"File {fpath2} does not exist")
		with open(fpath1, mode="rb") as file1:
			data1 = file1.read()
		with open(fpath2, mode="rb") as file2:
			data2 = file2.read()
		self.assertEqual(len(data1), len(data2), msg=f"{fpath1}")
		self.assertTrue(
			data1 == data2,
			msg=f"{fpath1} differs from {fpath2}",
		)

	def compareZipFiles(
		self,
		fpath1,
		fpath2,
		dataReplaceFuncs: dict[str, Callable],
	):
		zf1 = zipfile.ZipFile(fpath1)
		zf2 = zipfile.ZipFile(fpath2)
		pathList1 = zf1.namelist()
		pathList2 = zf2.namelist()
		if not self.assertEqual(pathList1, pathList2):
			return
		for zfpath in pathList1:
			data1 = zf1.read(zfpath)
			data2 = zf2.read(zfpath)

			func = dataReplaceFuncs.get(zfpath)
			if func is not None:
				data1 = func(data1)
				data2 = func(data2)

			self.assertEqual(len(data1), len(data2), msg=f"{zfpath=}")
			self.assertTrue(
				data1 == data2,
				msg=f"{zfpath=}",
			)

	def checkZipFileSha1sum(
		self,
		fpath,
		sha1sumDict: dict[str, str],
		dataReplaceFuncs: dict[str, Callable] | None = None,
	):
		if dataReplaceFuncs is None:
			dataReplaceFuncs = {}
		zf = zipfile.ZipFile(fpath)
		# pathList = zf.namelist()
		for zfpath, expectedSha1 in sha1sumDict.items():
			data = zf.read(zfpath)
			func = dataReplaceFuncs.get(zfpath)
			if func is not None:
				data = func(data)
			actualSha1 = hashlib.sha1(data).hexdigest()
			self.assertEqual(actualSha1, expectedSha1, msg=f"file: {zfpath}")

	def convert(  # noqa: PLR0913
		self,
		fname,  # input file with extension
		fname2,  # output file with extension
		testId="tmp",  # noqa: ARG002
		compareText="",
		compareBinary="",
		sha1sum=None,
		md5sum=None,
		config=None,
		showDiff=False,
		**convertArgs,
	):
		inputFilename = self.downloadFile(fname)
		outputFilename = self.newTempFilePath(fname2)
		glos = self.glos = Glossary()
		if config is not None:
			glos.config = config
		res = glos.convert(
			inputFilename=inputFilename,
			outputFilename=outputFilename,
			**convertArgs,
		)
		self.assertEqual(outputFilename, res)

		if compareText:
			self.compareTextFiles(
				outputFilename,
				self.downloadFile(compareText),
				showDiff=showDiff,
			)
		elif compareBinary:
			self.compareBinaryFiles(outputFilename, self.downloadFile(compareBinary))
		elif sha1sum:
			with open(outputFilename, mode="rb") as file:
				actualSha1 = hashlib.sha1(file.read()).hexdigest()
			self.assertEqual(actualSha1, sha1sum)

		elif md5sum:
			with open(outputFilename, mode="rb") as file:
				actualMd5 = hashlib.md5(file.read()).hexdigest()
			self.assertEqual(actualMd5, md5sum)

	def convert_sqlite_both(self, *args, **kwargs):
		for sqlite in (None, True, False):
			self.convert(*args, sqlite=sqlite, **kwargs)


class TestGlossary(TestGlossaryBase):
	def __init__(self, *args, **kwargs):
		TestGlossaryBase.__init__(self, *args, **kwargs)

		self.dataFileCRC32.update(
			{
				"100-en-fa-sort.txt": "d7a82dc8",
				"100-en-fa-sort-headword.txt": "4067a29f",
				"100-en-fa-sort-ebook.txt": "aa620d07",
				"100-en-fa-sort-ebook3.txt": "5a20f140",
				"100-en-fa-lower.txt": "62178940",
				"100-en-fa-remove_html_all-v3.txt": "d611c978",
				"100-en-fa-rtl.txt": "25ede1e8",
				"300-rand-en-fa-sort-headword-w1256.txt": "06d83bac",
				"300-rand-en-fa-sort-headword.txt": "df0f8020",
				"300-rand-en-fa-sort-w1256.txt": "9594aab3",
				"sort-locale/092-en-fa-alphabet-sample.txt": "b4856532",
				"sort-locale/092-en-fa-alphabet-sample-sorted-default.txt": "e7b70589",
				"sort-locale/092-en-fa-alphabet-sample-sorted-en.txt": "3d2bdf73",
				"sort-locale/092-en-fa-alphabet-sample-sorted-fa.txt": "245419db",
				"sort-locale/092-en-fa-alphabet-sample-sorted-latin-fa.txt": "261c03c0",
			},
		)

	def setUp(self):
		TestGlossaryBase.setUp(self)
		self.prevLogLevel = log.level
		log.setLevel(logging.ERROR)

	def tearDown(self):
		TestGlossaryBase.tearDown(self)
		log.setLevel(self.prevLogLevel)

	def test__str__1(self):
		glos = self.glos = Glossary()
		self.assertEqual(str(glos), "Glossary{filename: '', name: None}")

	def test__str__2(self):
		glos = self.glos = Glossary()
		glos._filename = "test.txt"
		self.assertEqual(str(glos), "Glossary{filename: 'test.txt', name: None}")

	def test__str__3(self):
		glos = self.glos = Glossary()
		glos.setInfo("title", "Test Title")
		self.assertEqual(
			str(glos),
			"Glossary{filename: '', name: 'Test Title'}",
		)

	def test__str__4(self):
		glos = self.glos = Glossary()
		glos._filename = "test.txt"
		glos.setInfo("title", "Test Title")
		self.assertEqual(
			str(glos),
			"Glossary{filename: 'test.txt', name: 'Test Title'}",
		)

	def test_info_1(self):
		glos = self.glos = Glossary()
		glos.setInfo("test", "ABC")
		self.assertEqual(glos.getInfo("test"), "ABC")

	def test_info_2(self):
		glos = self.glos = Glossary()
		glos.setInfo("bookname", "Test Glossary")
		self.assertEqual(glos.getInfo("title"), "Test Glossary")

	def test_info_3(self):
		glos = self.glos = Glossary()
		glos.setInfo("bookname", "Test Glossary")
		glos.setInfo("title", "Test 2")
		self.assertEqual(glos.getInfo("name"), "Test 2")
		self.assertEqual(glos.getInfo("bookname"), "Test 2")
		self.assertEqual(glos.getInfo("title"), "Test 2")

	def test_info_4(self):
		glos = self.glos = Glossary()
		glos.setInfo("test", 123)
		self.assertEqual(glos.getInfo("test"), "123")

	def test_info_del_1(self):
		glos = self.glos = Glossary()
		glos.setInfo("test", "abc")
		self.assertEqual(glos.getInfo("test"), "abc")
		glos.setInfo("test", None)
		self.assertEqual(glos.getInfo("test"), "")

	def test_info_del_2(self):
		glos = self.glos = Glossary()
		glos.setInfo("test", None)
		self.assertEqual(glos.getInfo("test"), "")

	def test_setInfo_err1(self):
		glos = self.glos = Glossary()
		try:
			glos.setInfo(1, "a")
		except TypeError as e:
			self.assertEqual(str(e), "invalid key=1, must be str")
		else:
			self.fail("must raise a TypeError")

	def test_getInfo_err1(self):
		glos = self.glos = Glossary()
		try:
			glos.getInfo(1)
		except TypeError as e:
			self.assertEqual(str(e), "invalid key=1, must be str")
		else:
			self.fail("must raise a TypeError")

	def test_getExtraInfos_1(self):
		glos = self.glos = Glossary()
		glos.setInfo("a", "test 1")
		glos.setInfo("b", "test 2")
		glos.setInfo("c", "test 3")
		glos.setInfo("d", "test 4")
		glos.setInfo("name", "my name")

		self.assertEqual(
			glos.getExtraInfos(["b", "c", "title"]),
			{"a": "test 1", "d": "test 4"},
		)

	def test_infoKeys_1(self):
		glos = self.glos = Glossary()
		glos.setInfo("a", "test 1")
		glos.setInfo("b", "test 2")
		glos.setInfo("name", "test name")
		glos.setInfo("title", "test title")

		self.assertEqual(
			glos.infoKeys(),
			["a", "b", "name"],
		)

	def test_config_attr_get(self):
		glos = self.glos = Glossary()
		try:
			glos.config  # noqa: B018
		except NotImplementedError:
			pass
		else:
			self.fail("must raise NotImplementedError")

	def test_config_attr_set(self):
		glos = self.glos = Glossary()
		glos.config = {"lower": True}
		self.assertEqual(glos.getConfig("lower", False), True)

	def test_read_txt_1(self):
		inputFilename = self.downloadFile("100-en-fa.txt")
		glos = self.glos = Glossary()
		res = glos.read(filename=inputFilename)
		self.assertTrue(res)
		self.assertEqual(glos.sourceLangName, "English")
		self.assertEqual(glos.targetLangName, "Persian")
		self.assertIn("Sample: ", glos.getInfo("name"))
		self.assertEqual(len(glos), 100)

	def test_read_txt_direct_1(self):
		inputFilename = self.downloadFile("100-en-fa.txt")
		glos = self.glos = Glossary()
		res = glos.read(filename=inputFilename, direct=True)
		self.assertTrue(res)
		self.assertEqual(glos.sourceLangName, "English")
		self.assertEqual(glos.targetLangName, "Persian")
		self.assertIn("Sample: ", glos.getInfo("name"))
		self.assertEqual(len(glos), 0)

	def test_init_infoDict(self):
		glos = self.glos = Glossary(info={"a": "b"})
		self.assertEqual(list(glos.iterInfo()), [("a", "b")])

	def test_init_infoOrderedDict(self):
		from collections import OrderedDict

		glos = self.glos = Glossary(
			info=OrderedDict(
				[
					("y", "z"),
					("a", "b"),
					("1", "2"),
				],
			),
		)
		self.assertEqual(list(glos.iterInfo()), [("y", "z"), ("a", "b"), ("1", "2")])

	def test_lang_1(self):
		glos = self.glos = Glossary()
		self.assertEqual(glos.sourceLangName, "")
		self.assertEqual(glos.targetLangName, "")
		glos.sourceLangName = "ru"
		glos.targetLangName = "de"
		self.assertEqual(glos.sourceLangName, "Russian")
		self.assertEqual(glos.targetLangName, "German")

	def test_lang_get_source(self):
		glos = self.glos = Glossary()
		glos.setInfo("sourcelang", "farsi")
		self.assertEqual(glos.sourceLangName, "Persian")

	def test_lang_get_target(self):
		glos = self.glos = Glossary()
		glos.setInfo("targetlang", "malay")
		self.assertEqual(glos.targetLangName, "Malay")

	def test_lang_set_source(self):
		glos = self.glos = Glossary()
		glos.sourceLangName = "en"
		self.assertEqual(glos.sourceLangName, "English")

	def test_lang_set_source_empty(self):
		glos = self.glos = Glossary()
		glos.sourceLangName = ""
		self.assertEqual(glos.sourceLangName, "")

	def test_lang_set_target(self):
		glos = self.glos = Glossary()
		glos.targetLangName = "fa"
		self.assertEqual(glos.targetLangName, "Persian")

	def test_lang_set_target_empty(self):
		glos = self.glos = Glossary()
		glos.targetLangName = ""
		self.assertEqual(glos.targetLangName, "")

	def test_lang_getObj_source(self):
		glos = self.glos = Glossary()
		glos.setInfo("sourcelang", "farsi")
		self.assertIsNotNone(glos.sourceLang)
		if glos.sourceLang is not None:
			self.assertEqual(glos.sourceLang.name, "Persian")

	def test_lang_getObj_target(self):
		glos = self.glos = Glossary()
		glos.setInfo("targetlang", "malay")
		self.assertIsNotNone(glos.targetLang)
		if glos.targetLang is not None:
			self.assertEqual(glos.targetLang.name, "Malay")

	def test_lang_detect_1(self):
		glos = self.glos = Glossary()
		glos.setInfo("name", "en-fa")
		glos.detectLangsFromName()
		self.assertEqual(
			(glos.sourceLangName, glos.targetLangName),
			("English", "Persian"),
		)

	def test_lang_detect_2(self):
		glos = self.glos = Glossary()
		glos.setInfo("name", "test-en-fa")
		glos.detectLangsFromName()
		self.assertEqual(
			(glos.sourceLangName, glos.targetLangName),
			("English", "Persian"),
		)

	def test_lang_detect_3(self):
		glos = self.glos = Glossary()
		glos.setInfo("name", "eng to per")
		glos.detectLangsFromName()
		self.assertEqual(
			(glos.sourceLangName, glos.targetLangName),
			("English", "Persian"),
		)

	def test_lang_detect_4(self):
		glos = self.glos = Glossary()
		glos.setInfo("name", "Test english to farsi")
		glos.detectLangsFromName()
		self.assertEqual(
			(glos.sourceLangName, glos.targetLangName),
			("English", "Persian"),
		)

	def test_lang_detect_5(self):
		glos = self.glos = Glossary()
		glos.setInfo("name", "freedict-eng-deu.index")
		glos.detectLangsFromName()
		self.assertEqual(
			(glos.sourceLangName, glos.targetLangName),
			("English", "German"),
		)

	def convert_txt_txt(
		self,
		fname,  # input txt file without extension
		fname2,  # expected output txt file without extension
		testId="tmp",
		config=None,
		**convertArgs,
	):
		self.convert(
			f"{fname}.txt",
			f"{fname2}-{testId}.txt",
			compareText=f"{fname2}.txt",
			testId=testId,
			config=config,
			**convertArgs,
		)

	def convert_to_txtZip(
		self,
		fname,  # input file with extension
		fname2,  # expected output file without extensions
		testId="tmp",
		config=None,
		**convertArgs,
	):
		inputFilename = self.downloadFile(fname)
		outputTxtName = f"{fname2}-{testId}.txt"
		outputFilename = self.newTempFilePath(f"{outputTxtName}.zip")
		expectedFilename = self.downloadFile(f"{fname2}.txt")
		glos = self.glos = Glossary()
		if config is not None:
			glos.config = config
		res = glos.convert(
			inputFilename=inputFilename,
			outputFilename=outputFilename,
			**convertArgs,
		)
		self.assertEqual(outputFilename, res)
		zf = zipfile.ZipFile(outputFilename)
		self.assertTrue(
			outputTxtName in zf.namelist(),
			msg=f"{outputTxtName} not in {zf.namelist()}",
		)
		with open(expectedFilename, encoding="utf-8") as expectedFile:
			expectedText = expectedFile.read()
		actualText = zf.read(outputTxtName).decode("utf-8")
		self.assertEqual(len(actualText), len(expectedText))
		self.assertEqual(actualText, expectedText)

	def test_txt_txtZip_1(self):
		self.convert_to_txtZip(
			"100-en-fa.txt",
			"100-en-fa",
			testId="txt_txtZip_1",
			infoOverride={"input_file_size": None},
		)

	def test_sort_1(self):
		self.convert_txt_txt(
			"100-en-fa",
			"100-en-fa-sort",
			testId="sort_1",
			sort=True,
		)

	def test_sort_2(self):
		self.convert_txt_txt(
			"100-en-fa",
			"100-en-fa-sort",
			testId="sort_2",
			sort=True,
			sortKeyName="headword_lower",
		)

	def test_sort_3(self):
		self.convert_txt_txt(
			"100-en-fa",
			"100-en-fa-sort-headword",
			testId="sort_3",
			sort=True,
			sortKeyName="headword",
		)

	def test_sort_4(self):
		self.convert_txt_txt(
			"300-rand-en-fa",
			"300-rand-en-fa-sort-headword",
			testId="sort_4",
			sort=True,
			sortKeyName="headword",
		)

	def test_sort_5(self):
		self.convert_txt_txt(
			"300-rand-en-fa",
			"300-rand-en-fa-sort-headword-w1256",
			testId="sort_5",
			sort=True,
			sortKeyName="headword",
			sortEncoding="windows-1256",
		)

	def test_sort_6(self):
		self.convert_txt_txt(
			"300-rand-en-fa",
			"300-rand-en-fa-sort-w1256",
			testId="sort_6",
			sort=True,
			sortKeyName="headword_lower",
			sortEncoding="windows-1256",
		)

	def test_sort_7(self):
		self.convert_txt_txt(
			"100-en-fa",
			"100-en-fa-sort-ebook",
			testId="sort_7",
			sort=True,
			sortKeyName="ebook",
		)

	def test_sort_8(self):
		self.convert_txt_txt(
			"100-en-fa",
			"100-en-fa-sort-ebook3",
			testId="sort_8",
			sort=True,
			sortKeyName="ebook_length3",
		)

	def test_lower_1(self):
		self.convert_txt_txt(
			"100-en-fa",
			"100-en-fa-lower",
			testId="lower_1",
			config={"lower": True},
		)

	def test_rtl_1(self):
		self.convert_txt_txt(
			"100-en-fa",
			"100-en-fa-rtl",
			testId="rtl_1",
			config={"rtl": True},
		)

	def test_remove_html_all_1(self):
		self.convert_txt_txt(
			"100-en-fa",
			"100-en-fa-remove_html_all-v3",
			testId="remove_html_all_1",
			config={"remove_html_all": True},
		)

	def test_remove_html_1(self):
		self.convert_txt_txt(
			"100-en-de-v4",
			"100-en-de-v4-remove_font_b",
			testId="remove_html_1",
			config={"remove_html": "font,b"},
		)

	def test_save_info_json(self):
		fname = "100-en-fa"
		testId = "save_info_json"
		infoPath = self.newTempFilePath(f"{fname}-{testId}.info")
		self.convert_txt_txt(
			fname,
			fname,
			testId=testId,
			config={"save_info_json": True},
			infoOverride={"input_file_size": None},
		)
		with open(infoPath, encoding="utf8") as file:
			infoDict = json.load(file)
		with open(self.downloadFile(f"{fname}-v2.info"), encoding="utf8") as file:
			infoDictExpected = json.load(file)
		for key, value in infoDictExpected.items():
			self.assertIn(key, infoDict)
			self.assertEqual(value, infoDict.get(key))

	def test_convert_sqlite_direct_error(self):
		glos = self.glos = Glossary()
		try:
			glos.convert(
				inputFilename="foo.txt",
				outputFilename="bar.txt",
				direct=True,
				sqlite=True,
			)
		except ValueError as e:
			self.assertEqual(str(e), "Conflictng arguments: direct=True, sqlite=True")
		else:
			self.fail("must raise a ValueError")

	def test_txt_txt_bar(self):
		for direct in (None, False, True):
			self.convert_txt_txt(
				"004-bar",
				"004-bar",
				testId="bar",
				direct=direct,
				infoOverride={
					"name": None,
					"input_file_size": None,
				},
			)

	def test_txt_txt_bar_sort(self):
		for sqlite in (None, False, True):
			self.convert_txt_txt(
				"004-bar",
				"004-bar-sort",
				testId="bar_sort",
				sort=True,
				sqlite=sqlite,
			)

	def test_txt_txt_empty_filtered(self):
		for direct in (None, False, True):
			self.convert_txt_txt(
				"006-empty",
				"006-empty-filtered",
				testId="empty_filtered",
				direct=direct,
			)

	def test_txt_txt_empty_filtered_sqlite(self):
		for sqlite in (None, False, True):
			self.convert_txt_txt(
				"006-empty",
				"006-empty-filtered",
				testId="empty_filtered_sqlite",
				sqlite=sqlite,
			)

	def test_dataEntry_save(self):
		glos = self.glos = Glossary()
		tmpFname = "test_dataEntry_save"
		entry = glos.newDataEntry(tmpFname, b"test")
		saveFpath = entry.save(self.tempDir)
		self.assertTrue(
			isfile(saveFpath),
			msg=f"saved file does not exist: {saveFpath}",
		)

	def test_dataEntry_getFileName(self):
		glos = self.glos = Glossary()
		tmpFname = "test_dataEntry_getFileName"
		entry = glos.newDataEntry(tmpFname, b"test")
		self.assertEqual(entry.getFileName(), tmpFname)

	def test_cleanup_noFile(self):
		glos = self.glos = Glossary()
		glos.cleanup()

	def test_cleanup_cleanup(self):
		glos = self.glos = Glossary()
		tmpFname = "test_cleanup_cleanup"
		entry = glos.newDataEntry(tmpFname, b"test")

		tmpFpath = entry.tmpPath
		self.assertTrue(bool(tmpFpath), msg="entry tmpPath is empty")
		if tmpFpath:
			self.assertTrue(
				isfile(tmpFpath),
				msg=f"tmp file does not exist: {tmpFpath}",
			)

		glos.cleanup()

		if tmpFpath:
			self.assertTrue(
				not isfile(tmpFpath),
				msg=f"tmp file still exists: {tmpFpath}",
			)

	def test_cleanup_noCleanup(self):
		glos = self.glos = Glossary()
		tmpFname = "test_cleanup_noCleanup"
		entry = glos.newDataEntry(tmpFname, b"test")

		tmpFpath = entry.tmpPath
		self.assertTrue(bool(tmpFpath), msg="entry tmpPath is empty")
		if tmpFpath:
			self.assertTrue(
				isfile(tmpFpath),
				msg=f"tmp file does not exist: {tmpFpath}",
			)

		glos.config = {"cleanup": False}
		glos.cleanup()

		if tmpFpath:
			self.assertTrue(
				isfile(tmpFpath),
				msg=f"tmp file does not exist: {tmpFpath}",
			)

	def addWordsList(
		self,
		glos: Glossary,
		terms: list[str],
		newDefiFunc: Callable[[Any], str] = str,
		defiFormat: str = "",
	) -> list[list[str]]:
		wordsList = []
		for index, line in enumerate(terms):
			terms = line.rstrip().split("|")
			wordsList.append(terms)
			glos.addEntryObj(
				glos.newEntry(
					terms,
					newDefiFunc(index),
					defiFormat=defiFormat,
				),
			)

		glos.updateIter()
		return wordsList

	def addWords(self, glos, wordsStr, **kwargs):
		return self.addWordsList(glos, wordsStr.split("\n"), **kwargs)

	tenWordsStr = """comedic
tubenose
organosol
adipocere
gid
next friend
bitter apple
caca|ca-ca
darkling beetle
japonica"""

	tenWordsStr2 = """comedic
Tubenose
organosol
Adipocere
gid
Next friend
bitter apple
Caca|ca-ca
darkling beetle
Japonica"""

	tenWordsStrFa = (
		"بیمارانه\nگالوانومتر\nنقاهت\nرشک"
		"مندی\nناکاستنی\nشگفتآفرینی\nچندپاری\nنامبارکی\nآماسش\nانگیزنده"
	)

	def test_addEntries_1(self):
		glos = self.glos = Glossary()
		wordsList = self.addWords(
			glos,
			self.tenWordsStr,
			newDefiFunc=lambda _i: str(random.randint(0, 10000)),
		)
		self.assertEqual(wordsList, [entry.l_term for entry in glos])

	def test_addEntries_2(self):
		# entry filters don't apply to loaded entries (added with addEntryObj)
		glos = self.glos = Glossary()
		glos.addEntryObj(glos.newEntry(["a"], "test 1"))
		glos.addEntryObj(glos.newEntry([""], "test 2"))
		glos.addEntryObj(glos.newEntry(["b"], "test 3"))
		glos.addEntryObj(glos.newEntry([], "test 4"))
		glos.updateEntryFilters()
		glos.updateIter()
		self.assertEqual(
			[["a"], [""], ["b"], []],
			[entry.l_term for entry in glos],
		)

	def test_addEntries_3(self):
		glos = self.glos = Glossary()
		glos.addEntryObj(glos.newEntry(["a"], "test 1"))
		glos.addEntryObj(glos.newEntry(["b"], "test 3"))
		glos.addEntryObj(
			glos.newDataEntry(
				"file.bin",
				b"hello\x00world",
			),
		)
		glos.updateEntryFilters()
		glos.updateIter()
		wordListList = []
		dataEntries = []
		for entry in glos:
			wordListList.append(entry.l_term)
			if entry.isData():
				dataEntries.append(entry)
		self.assertEqual(
			wordListList,
			[["a"], ["b"], ["file.bin"]],
		)
		self.assertEqual(len(dataEntries), 1)
		self.assertEqual(dataEntries[0].getFileName(), "file.bin")
		self.assertEqual(dataEntries[0].data, b"hello\x00world")

	def test_sortWords_1(self):
		glos = self.glos = Glossary()
		wordsList = self.addWords(
			glos,
			self.tenWordsStr,
			newDefiFunc=lambda _i: str(random.randint(0, 10000)),
		)
		self.assertEqual(wordsList, [entry.l_term for entry in glos])
		glos.sortWords()
		self.assertEqual(sorted(wordsList), [entry.l_term for entry in glos])

	def test_sortWords_2(self):
		glos = self.glos = Glossary()
		wordsList = self.addWords(
			glos,
			self.tenWordsStr2,
			newDefiFunc=lambda _i: str(random.randint(0, 10000)),
		)
		self.assertEqual(wordsList, [entry.l_term for entry in glos])
		glos.sortWords(sortKeyName="headword")
		self.assertEqual(
			[entry.l_term for entry in glos],
			[
				["Adipocere"],
				["Caca", "ca-ca"],
				["Japonica"],
				["Next friend"],
				["Tubenose"],
				["bitter apple"],
				["comedic"],
				["darkling beetle"],
				["gid"],
				["organosol"],
			],
		)

	def test_sortWords_3(self):
		glos = self.glos = Glossary()
		wordsList = self.addWords(
			glos,
			self.tenWordsStrFa,
			newDefiFunc=lambda _i: str(random.randint(0, 10000)),
		)
		self.assertEqual(wordsList, [entry.l_term for entry in glos])
		glos.sortWords(sortKeyName="headword")
		ls1 = ["آماسش", "انگیزنده", "بیمارانه", "رشکمندی", "شگفتآفرینی"]
		ls2 = ["نامبارکی", "ناکاستنی", "نقاهت", "چندپاری", "گالوانومتر"]
		self.assertEqual(
			[entry.s_term for entry in glos],
			ls1 + ls2,
		)

	def test_sortWords_4(self):
		glos = self.glos = Glossary()
		wordsList = self.addWords(
			glos,
			self.tenWordsStrFa,
			newDefiFunc=lambda _i: str(random.randint(0, 10000)),
		)
		self.assertEqual(wordsList, [entry.l_term for entry in glos])
		glos.sortWords(
			sortKeyName="headword",
			sortEncoding="windows-1256",
		)
		ls1 = ["چندپاری", "گالوانومتر", "آماسش", "انگیزنده", "بیمارانه"]
		ls2 = ["رشکمندی", "شگفتآفرینی", "ناکاستنی", "نامبارکی", "نقاهت"]
		self.assertEqual(
			[entry.s_term for entry in glos],
			ls1 + ls2,
		)

	def test_sortWords_5(self):
		glos = self.glos = Glossary()
		alphabetW1256 = "ءآأئابتثجحخدذرزسشصضطظعغـفقكلمنهوىي"
		alphabetW1256_shuf = "مفزنصـذرخوآظسقلدغطيعحءأتىئاجهضثشكب"
		wordsList = self.addWordsList(
			glos,
			list(alphabetW1256_shuf),
			newDefiFunc=lambda _x: str(random.randint(0, 10000)),
		)
		self.assertEqual(wordsList, [entry.l_term for entry in glos])
		glos.sortWords(
			sortKeyName="headword",
			sortEncoding="windows-1256",
		)
		self.assertEqual(
			[entry.s_term for entry in glos],
			list(alphabetW1256),
		)

	def test_sortWords_exc_1(self):
		fname = "100-en-fa.txt"
		glos = self.glos = Glossary()
		glos.read(self.downloadFile(fname), direct=True)
		try:
			glos.sortWords()
		except NotImplementedError as e:
			self.assertEqual(str(e), "can not use sortWords in direct mode")
		else:
			self.fail("must raise NotImplementedError")

	def test_read_filename(self):
		glos = self.glos = Glossary()
		glos.read(self.downloadFile("004-bar.txt"))
		self.assertEqual(glos.filename, join(testCacheDir, "004-bar"))

	def test_wordTitleStr_em1(self):
		glos = self.glos = Glossary()
		self.assertEqual(glos.wordTitleStr(""), "")

	def test_wordTitleStr_em2(self):
		glos = self.glos = Glossary()
		glos._defiHasWordTitle = True
		self.assertEqual(glos.wordTitleStr("test1"), "")

	def test_wordTitleStr_b1(self):
		glos = self.glos = Glossary()
		self.assertEqual(glos.wordTitleStr("test1"), "<b>test1</b><br>")

	def test_wordTitleStr_b2(self):
		glos = self.glos = Glossary()
		self.assertEqual(
			glos.wordTitleStr("test1", class_="headword"),
			'<b class="headword">test1</b><br>',
		)

	def test_wordTitleStr_cjk1(self):
		glos = self.glos = Glossary()
		self.assertEqual(
			glos.wordTitleStr("test1", sample="くりかえし"),
			"<big>test1</big><br>",
		)

	def test_wordTitleStr_cjk2(self):
		glos = self.glos = Glossary()
		self.assertEqual(
			glos.wordTitleStr("くりかえし"),
			"<big>くりかえし</big><br>",
		)

	def test_convert_sortLocale_default_1(self):
		name = "092-en-fa-alphabet-sample"
		self.convert_sqlite_both(
			f"sort-locale/{name}.txt",
			f"{name}-sorted-default.txt",
			compareText=f"sort-locale/{name}-sorted-default.txt",
			testId="sorted-default",
			sort=True,
			sortKeyName="headword_lower",
		)

	def test_convert_sortLocale_en_1(self):
		name = "092-en-fa-alphabet-sample"
		self.convert_sqlite_both(
			f"sort-locale/{name}.txt",
			f"{name}-sorted-en.txt",
			compareText=f"sort-locale/{name}-sorted-en.txt",
			testId="sorted-en",
			sort=True,
			sortKeyName="headword_lower:en_US.UTF-8",
		)

	def test_convert_sortLocale_fa_1(self):
		name = "092-en-fa-alphabet-sample"
		self.convert_sqlite_both(
			f"sort-locale/{name}.txt",
			f"{name}-sorted-fa.txt",
			compareText=f"sort-locale/{name}-sorted-fa.txt",
			testId="sorted-fa",
			sort=True,
			sortKeyName="headword_lower:fa_IR.UTF-8",
		)

	def test_convert_sortLocale_fa_2(self):
		name = "092-en-fa-alphabet-sample"
		self.convert_sqlite_both(
			f"sort-locale/{name}.txt",
			f"{name}-sorted-latin-fa.txt",
			compareText=f"sort-locale/{name}-sorted-latin-fa.txt",
			testId="sorted-latin-fa",
			sort=True,
			sortKeyName="headword_lower:fa-u-kr-latn-arab",
		)


if __name__ == "__main__":
	unittest.main()



import gzip
import logging
import unittest
from pathlib import Path

from glossary_v2_errors_test import TestGlossaryErrorsBase

from pyglossary.os_utils import runDictzip

TEXT = """
Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor
incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis
nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat.
Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu
fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in
culpa qui officia deserunt mollit anim id est laborum.
"""
MISSING_DEP_MARK = "Dictzip compression requires idzip module or dictzip utility,"


class TestDictzip(TestGlossaryErrorsBase):
	def setUp(self) -> None:
		super().setUp()
		self.test_file_path = Path(self.tempDir) / "test_file.txt"
		filename = self.test_file_path.name + ".dz"
		self.result_file_path = self.test_file_path.parent / filename
		with open(self.test_file_path, "a", encoding="utf-8") as tmp_file:
			tmp_file.write(TEXT)

	def skip_on_dep(self, method: str) -> None:
		warn = self.mockLog.popLog(logging.WARNING, MISSING_DEP_MARK, partial=True)
		if warn:
			self.skipTest(f"Missing {method} dependency")

	def test_idzip_compressed_exists(self) -> None:
		method = "idzip"
		runDictzip(self.test_file_path, method)
		self.skip_on_dep(method)
		self.assertTrue(self.result_file_path.exists())
		self.assertTrue(self.result_file_path.is_file())

	def test_idzip_compressed_matches(self) -> None:
		method = "idzip"
		runDictzip(self.test_file_path, method)
		self.skip_on_dep(method)
		with gzip.open(self.result_file_path, "r") as file:
			result = file.read().decode()
		self.assertEqual(result, TEXT)

	def test_dictzip_compressed_exists(self) -> None:
		method = "dictzip"
		runDictzip(self.test_file_path, method)
		self.skip_on_dep(method)
		self.assertTrue(self.result_file_path.exists())
		self.assertTrue(self.result_file_path.is_file())

	def test_dictzip_compressed_matches(self) -> None:
		method = "dictzip"
		runDictzip(self.test_file_path, method)
		self.skip_on_dep(method)
		with gzip.open(self.result_file_path, "r") as file:
			result = file.read().decode()
		self.assertEqual(result, TEXT)

	def test_dictzip_missing_target(self) -> None:
		method = "idzip"
		filename = "/NOT_EXISTED_PATH/file.txt"
		expected = f"No such file or directory: '{filename}'"
		runDictzip(filename, method)
		self.skip_on_dep(method)
		err = self.mockLog.popLog(logging.ERROR, expected, partial=True)
		self.assertIsNotNone(err)

	def test_idzip_missing_target(self) -> None:
		method = "dictzip"
		filename = "/NOT_EXISTED_PATH/boilerplate.txt"
		expected = f'Cannot open "{filename}"'
		runDictzip(filename, method)
		self.skip_on_dep(method)
		err = self.mockLog.popLog(logging.ERROR, expected, partial=True)
		self.assertIsNotNone(err)


if __name__ == "__main__":
	unittest.main()



# -*- coding: utf-8 -*-
# mypy: ignore-errors
#
# Copyright © 2025 Saeed Rasooli <saeed.gnu@gmail.com> (ilius)
#
# This program is a free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
#
# You can get a copy of GNU General Public License along this program
# But you can always get it from http://www.gnu.org/licenses/gpl.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from gi.repository import Gtk as gtk

from pyglossary.ui.config import configDefDict

from .sort_options import SortOptionsBox
from .utils import (
	HBox,
	pack,
)

if TYPE_CHECKING:
	from gi.repository import Gdk as gdk

__all__ = ["GeneralOptionsDialog"]


class GeneralOptionsDialog(gtk.Dialog):
	def onCloseRequest(self, _widget: gtk.Widget) -> bool:
		self.hide()
		return True

	def onResponse(self, _widget: gtk.Widget, _event: gdk.Event) -> bool:
		self.applyChanges()
		self.hide()
		return True

	def __init__(self, mainWin: gtk.Window, **kwargs: Any) -> None:
		gtk.Dialog.__init__(
			self,
			transient_for=mainWin,
			**kwargs,
		)
		self.set_title("General Options")
		self.mainWin = mainWin
		##
		self.vbox = self.get_content_area()
		self.vbox.set_spacing(5)
		##
		self.set_default_size(600, 500)
		self.connect("close-request", self.onCloseRequest)
		##
		self.connect("response", self.onResponse)
		self.add_action_widget(
			gtk.Button(
				label="_OK",
				use_underline=True,
				# icon_name="gtk-ok",
			),
			gtk.ResponseType.OK,
		)
		##
		hpad = 10
		##
		self.sortOptionsBox = SortOptionsBox(mainWin)
		pack(self.vbox, self.sortOptionsBox)
		##
		hbox = HBox(spacing=hpad)
		self.sqliteCheck = gtk.CheckButton(label="SQLite mode")
		pack(hbox, self.sqliteCheck)
		pack(self.vbox, hbox)
		##
		self.configParams = {
			"save_info_json": False,
			"lower": False,
			"skip_resources": False,
			"rtl": False,
			"enable_alts": True,
			"cleanup": True,
			"remove_html_all": True,
		}
		self.configCheckButtons = {}
		for param in self.configParams:
			hbox = HBox(spacing=hpad)
			comment = configDefDict[param].comment
			checkButton = gtk.CheckButton(
				label=comment.split("\n")[0],
			)
			self.configCheckButtons[param] = checkButton
			pack(hbox, checkButton)
			pack(self.vbox, hbox)
		##
		self.updateWidgets()
		self.vbox.show()

	def getSQLite(self) -> bool:
		convertOptions = self.mainWin.convertOptions
		sqlite = convertOptions.get("sqlite")
		if sqlite is not None:
			return sqlite
		return self.mainWin.config.get("auto_sqlite", True)

	def updateWidgets(self) -> None:
		config = self.mainWin.config
		self.sortOptionsBox.updateWidgets()
		self.sqliteCheck.set_active(self.getSQLite())
		for param, check in self.configCheckButtons.items():
			default = self.configParams[param]
			check.set_active(config.get(param, default))

	def applyChanges(self) -> None:
		# print("applyChanges")
		self.sortOptionsBox.applyChanges()

		convertOptions = self.mainWin.convertOptions
		config = self.mainWin.config

		convertOptions["sqlite"] = self.sqliteCheck.get_active()

		for param, check in self.configCheckButtons.items():
			config[param] = check.get_active()



import datetime
import hashlib
import os
import re
import sys
import unittest
from os.path import abspath, dirname

from freezegun import freeze_time

rootDir = dirname(dirname(abspath(__file__)))
sys.path.insert(0, rootDir)

from glossary_v2_test import TestGlossaryBase

from pyglossary.glossary_v2 import ConvertArgs, Glossary

testTimeEpoch = 1730579400
testTime = datetime.datetime.fromtimestamp(testTimeEpoch, tz=datetime.UTC)


class TestGlossaryEPUB2(TestGlossaryBase):
	def __init__(self, *args, **kwargs):
		TestGlossaryBase.__init__(self, *args, **kwargs)

		self.dataFileCRC32.update(
			{
				"100-en-fa-res.slob": "0216d006",
				"100-en-fa-res-slob-v2.epub": "304d174d",
				"100-en-fa-prefix3-v2.epub": "1b7244ca",
				"300-rand-en-fa-prefix3-v2.epub": "b5dd9ec6",
			},
		)

	def setUp(self):
		TestGlossaryBase.setUp(self)

	def remove_toc_uid(self, data):
		return re.sub(
			b'<meta name="dtb:uid" content="[0-9a-f]{32}" />',
			b'<meta name="dtb:uid" content="" />',
			data,
		)

	def remove_content_extra(self, data):
		data = re.sub(
			b'<dc:identifier id="uid" opf:scheme="uuid">[0-9a-f]{32}</dc:identifier>',
			b'<dc:identifier id="uid" opf:scheme="uuid"></dc:identifier>',
			data,
		)
		return re.sub(
			b'<dc:date opf:event="creation">[0-9-]{10}</dc:date>',
			b'<dc:date opf:event="creation"></dc:date>',
			data,
		)

	def convert_to_epub(
		self,
		inputFname,
		outputFname,
		testId,
		checkZipContents=True,
		sha1sum="",
		**convertArgs,
	):
		inputFilename = self.downloadFile(f"{inputFname}")
		outputFilename = self.newTempFilePath(
			f"{inputFname.replace('.', '_')}-{testId}.epub",
		)

		if sha1sum:
			os.environ["EPUB_UUID"] = hashlib.sha1(inputFname.encode("ascii")).hexdigest()
			os.environ["EBOOK_CREATION_TIME"] = str(testTimeEpoch)
			# print(f'{os.environ["EPUB_UUID"]=}')

		glos = self.glos = Glossary()
		res = glos.convert(
			ConvertArgs(
				inputFilename=inputFilename,
				outputFilename=outputFilename,
				**convertArgs,
			)
		)
		self.assertEqual(outputFilename, res)

		if checkZipContents:
			self.compareZipFiles(
				outputFilename,
				self.downloadFile(f"{outputFname}.epub"),
				{
					"OEBPS/toc.ncx": self.remove_toc_uid,
					"OEBPS/content.opf": self.remove_content_extra,
				},
			)
		if sha1sum:
			with open(outputFilename, mode="rb") as file:
				actualSha1 = hashlib.sha1(file.read()).hexdigest()
			self.assertEqual(sha1sum, actualSha1, f"{outputFilename=}")

	# sha1sum still depends on current date (but not time)
	# despite using ReproducibleZipFile and EBOOK_CREATION_TIME env var. not sure why.
	@freeze_time(testTime)
	def test_convert_txt_epub_1(self):
		self.convert_to_epub(
			"100-en-fa.txt",
			"100-en-fa",
			testId="a1",
			checkZipContents=False,
			sha1sum="beb3ad8227dd9561223cd18d805aff8dd0aef27b",
		)

	def test_convert_to_epub_1(self):
		self.convert_to_epub(
			"100-en-fa-res.slob",
			"100-en-fa-res-slob-v2",
			testId="1",
		)

	def test_convert_to_epub_2(self):
		for sort in (True, False):
			self.convert_to_epub(
				"100-en-fa-res.slob",
				"100-en-fa-res-slob-v2",
				testId="2",
				sort=sort,
			)

	def test_convert_to_epub_3(self):
		for sqlite in (True, False):
			self.convert_to_epub(
				"100-en-fa-res.slob",
				"100-en-fa-res-slob-v2",
				testId="3",
				sqlite=sqlite,
			)

	def test_convert_to_epub_4(self):
		for direct in (True, False):
			self.convert_to_epub(
				"100-en-fa-res.slob",
				"100-en-fa-res-slob-v2",
				testId="4",
				direct=direct,
			)

	def test_convert_to_epub_5(self):
		for sqlite in (True, False):
			self.convert_to_epub(
				"100-en-fa.txt",
				"100-en-fa-prefix3-v2",
				testId="5",
				sqlite=sqlite,
				writeOptions={"group_by_prefix_length": 3},
			)

	def test_convert_to_epub_6(self):
		self.convert_to_epub(
			"300-rand-en-fa.txt",
			"300-rand-en-fa-prefix3-v2",
			testId="6",
			sqlite=True,
			writeOptions={"group_by_prefix_length": 3},
		)

	def test_convert_to_epub_7(self):
		self.convert_to_epub(
			"300-rand-en-fa.txt",
			"300-rand-en-fa-prefix3-v2",
			testId="7",
			sqlite=False,
			writeOptions={"group_by_prefix_length": 3},
		)


if __name__ == "__main__":
	unittest.main()



import sys
import unittest
from os.path import abspath, dirname

rootDir = dirname(dirname(abspath(__file__)))
sys.path.insert(0, rootDir)


from g_stardict_test import TestGlossaryStarDictBase
from glossary_v2_errors_test import TestGlossaryErrorsBase

__all__ = ["TestGlossaryStarDictMergeSyns"]


class TestGlossaryStarDictMergeSyns(TestGlossaryStarDictBase):
	def convert_txt_stardict(self, *args, **kwargs):
		kwargs["outputFormat"] = "StardictMergeSyns"
		TestGlossaryStarDictBase.convert_txt_stardict(self, *args, **kwargs)

	def __init__(self, *args, **kwargs):
		TestGlossaryErrorsBase.__init__(self, *args, **kwargs)

		self.dataFileCRC32.update(
			{
				"002-plain-html.txt": "75484314",
				"004-plain-html-alts.txt": "505d4675",
				"002-plain-html-sd-merge-syns-v2/002-plain-html.dict": "2e9d20d8",
				"002-plain-html-sd-merge-syns-v2/002-plain-html.idx": "3956ad72",
				"002-plain-html-sd-merge-syns-v2/002-plain-html.ifo": "1991f125",
				"004-plain-html-alts-sd-merge-syns-v2/004-plain-html-alts.dict": "889f11f8",
				"004-plain-html-alts-sd-merge-syns-v2/004-plain-html-alts.idx": "092ba555",
				"004-plain-html-alts-sd-merge-syns-v2/004-plain-html-alts.ifo": "628abe99",
				# "004-plain-html-alts-sd-merge-syns-v2/004-plain-html-alts.syn": "c07f7111",
				"100-en-de-v4-sd-merge-syns-v2/100-en-de-v4.dict": "5a97476f",
				"100-en-de-v4-sd-merge-syns-v2/100-en-de-v4.idx": "a99f29d2",
				"100-en-de-v4-sd-merge-syns-v2/100-en-de-v4.ifo": "2120708c",
				"100-en-fa-sd-merge-syns-v2/100-en-fa.dict": "223a0d1d",
				"100-en-fa-sd-merge-syns-v2/100-en-fa.idx": "13f1c7af",
				"100-en-fa-sd-merge-syns-v2/100-en-fa.ifo": "248ef828",
			},
		)

	def test_convert_txt_stardict_1_merge_syns(self):
		self.convert_txt_stardict(
			"100-en-fa",
			"100-en-fa-sd-merge-syns-v2",
			syn=False,
			# dictzip=False,
		)

	def test_convert_txt_stardict_3_merge_syns(self):
		self.convert_txt_stardict(
			"100-en-de-v4",
			"100-en-de-v4-sd-merge-syns-v2",
			syn=False,
			# dictzip=False,
		)

	def test_convert_txt_stardict_general_1_merge_syns(self):
		self.convert_txt_stardict(
			"002-plain-html",
			"002-plain-html-sd-merge-syns-v2",
			syn=False,
			# dictzip=False,
		)

	def test_convert_txt_stardict_general_2_merge_syns(self):
		self.convert_txt_stardict(
			"004-plain-html-alts",
			"004-plain-html-alts-sd-merge-syns-v2",
			syn=False,
			# dictzip=False,
		)


if __name__ == "__main__":
	unittest.main()



import unittest

from glossary_v2_test import TestGlossaryBase


class TestGlossaryXDXFLax(TestGlossaryBase):
	def __init__(self, *args, **kwargs):
		TestGlossaryBase.__init__(self, *args, **kwargs)

		self.dataFileCRC32.update(
			{
				"100-cyber_lexicon_en-es.xdxf": "8d9ba394",
				"100-cyber_lexicon_en-es-v3.txt": "4aa05086",
			},
		)

	def convert_xdxf_txt(self, fname, fname2, **convertArgs):
		self.convert(
			f"{fname}.xdxf",
			f"{fname}-tmp.txt",
			compareText=f"{fname2}.txt",
			inputFormat="XdxfLax",
			**convertArgs,
		)

	def test_convert_xdxf_txt_1(self):
		self.convert_xdxf_txt(
			"100-cyber_lexicon_en-es",
			"100-cyber_lexicon_en-es-v3",
		)


if __name__ == "__main__":
	unittest.main()



import unittest

import mistune  # noqa: F401, to ensure it's installed
from glossary_v2_test import TestGlossaryBase


class TestGlossaryDictfile(TestGlossaryBase):
	def __init__(self, *args, **kwargs):
		TestGlossaryBase.__init__(self, *args, **kwargs)

		self.dataFileCRC32.update(
			{
				"022-en-en.df": "edff6de1",
				"022-en-en.df.txt": "93a2450f",
				"022-en-en.df.txt.df": "8e952e56",
				"res/01cf5b41.gif": "01cf5b41",
				"res/1f3c1a36.gif": "1f3c1a36",
				"res/3af9fd5d.gif": "3af9fd5d",
				"res/6684158d.gif": "6684158d",
			},
		)

	def convert_df_txt(self, fname, fname2, resFiles, **convertArgs):
		resFilesPath = {
			resFileName: self.newTempFilePath(f"{fname}-2.txt_res/{resFileName}")
			for resFileName in resFiles
		}

		self.convert(
			f"{fname}.df",
			f"{fname}-2.txt",
			compareText=f"{fname2}.txt",
			**convertArgs,
		)

		for resFileName in resFiles:
			fpath1 = self.downloadFile(f"res/{resFileName}")
			fpath2 = resFilesPath[resFileName]
			self.compareBinaryFiles(fpath1, fpath2)

	def convert_txt_df(self, fname, fname2, **convertArgs):
		self.convert(
			f"{fname}.txt",
			f"{fname}-2.df",
			compareText=f"{fname2}.df",
			**convertArgs,
		)

	def test_convert_df_txt_1(self):
		self.convert_df_txt(
			"022-en-en",
			"022-en-en.df",
			resFiles=[
				"01cf5b41.gif",
				"1f3c1a36.gif",
				"3af9fd5d.gif",
				"6684158d.gif",
			],
		)

	def test_convert_txt_df_1(self):
		self.convert_txt_df(
			"022-en-en.df",
			"022-en-en.df.txt",
		)


if __name__ == "__main__":
	unittest.main()



import re
from collections.abc import Generator
from typing import NamedTuple

__all__ = ["get_chinese_references"]


class ChineseWordReference(NamedTuple):
	text: str
	trad: str
	simp: str
	start: int
	end: int
	trad_start: int
	trad_end: int
	simp_start: int
	simp_end: int


def get_chinese_references(text: str) -> Generator[ChineseWordReference]:
	# Matches Chinese characters including A-Z, 0-9,
	# and full/half-width punctuation.
	han = r"[\u4E00-\u9FFF\uFF00-\uFFEFA-Z0-9]"

	# Matches trad|simp[pinyin] or both[pinyin].
	chinese_refs = rf"((?P<trad>{han}+)\|(?P<simp>{han}+)|(?P<both>{han}+))\[[^\]]+\]"
	chinese_matches = re.finditer(chinese_refs, text)

	for match in chinese_matches:
		if match.group("both") is not None:
			yield ChineseWordReference(
				text=match.group(0),
				trad=match.group("both"),
				simp=match.group("both"),
				start=match.start(),
				end=match.end(),
				trad_start=match.start("both"),
				trad_end=match.end("both"),
				simp_start=match.start("both"),
				simp_end=match.end("both"),
			)
		else:
			yield ChineseWordReference(
				text=match.group(0),
				trad=match.group("trad"),
				simp=match.group("simp"),
				start=match.start(),
				end=match.end(),
				trad_start=match.start("trad"),
				trad_end=match.end("trad"),
				simp_start=match.start("simp"),
				simp_end=match.end("simp"),
			)
